/*
   3D transformation matrix definition and routines.

   clib v1.1

   Copyright (C) 1997-1998 Per Kraulis
    27-Dec-1996  first attempts
    29-Jan-1997  fairly finished
    17-Jun-1998  mod's for hgen
    28-Jan-2002  TimF - add invert, diagonalize, aniso B P(C) table
*/

#include "matrix3.h"

/* public ====================
#include <vector3.h>
==================== public */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <math.h>

/*------------------------------------------------------------*/
void
matrix3_initialize (double m[4][4])
     /*
       Initialize to the unit matrix.
     */
{
  /* pre */
  assert (m);

  m[0][0] = 1.0;
  m[0][1] = 0.0;
  m[0][2] = 0.0;
  m[0][3] = 0.0;
  m[1][0] = 0.0;
  m[1][1] = 1.0;
  m[1][2] = 0.0;
  m[1][3] = 0.0;
  m[2][0] = 0.0;
  m[2][1] = 0.0;
  m[2][2] = 1.0;
  m[2][3] = 0.0;
  m[3][0] = 0.0;
  m[3][1] = 0.0;
  m[3][2] = 0.0;
  m[3][3] = 1.0;
}


/*------------------------------------------------------------*/
void
matrix3_copy (double dest[4][4], double src[4][4])
{
  /* pre */
  assert (dest);
  assert (src);

  memcpy (dest, src, 16 * sizeof (double));
}


/*------------------------------------------------------------*/
void
matrix3_concatenate (double dest[4][4], double src[4][4])
{
  register int i, j;
  double t[4][4];

  /* pre */
  assert (dest);
  assert (src);

  memcpy (t, dest, 16 * sizeof (double));

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      dest[i][j] = t[i][0] * src[0][j] +
	           t[i][1] * src[1][j] +
	           t[i][2] * src[2][j] +
	           t[i][3] * src[3][j];
    }
  }
}


/*------------------------------------------------------------*/
void
matrix3_x_rotation (double m[4][4], double radians)
{
  register double d;

  /* pre */
  assert (m);

  matrix3_initialize (m);

  d = cos (radians);
  m[1][1] = d;
  m[2][2] = d;
  d = sin (radians);
  m[1][2] = d;
  m[2][1] = -d;
}


/*------------------------------------------------------------*/
void
matrix3_y_rotation (double m[4][4], double radians)
{
  register double d;

  /* pre */
  assert (m);

  matrix3_initialize (m);

  d = cos (radians);
  m[0][0] = d;
  m[2][2] = d;
  d = sin (radians);
  m[2][0] = d;
  m[0][2] = -d;
}


/*------------------------------------------------------------*/
void
matrix3_z_rotation (double m[4][4], double radians)
{
  register double d;

  /* pre */
  assert (m);

  matrix3_initialize (m);

  d = cos (radians);
  m[0][0] = d;
  m[1][1] = d;
  d = sin (radians);
  m[0][1] = d;
  m[1][0] = -d;
}


/*------------------------------------------------------------*/
void
matrix3_translation (double m[4][4], double x, double y, double z)
{
  assert (m);

  /* pre */
  matrix3_initialize (m);

  m[3][0] = x;
  m[3][1] = y;
  m[3][2] = z;
}


/*------------------------------------------------------------*/
void
matrix3_scale (double m[4][4], double sx, double sy, double sz)
{
  assert (m);

  /* pre */
  matrix3_initialize (m);

  m[0][0] = sx;
  m[1][1] = sy;
  m[2][2] = sz;
}


/*------------------------------------------------------------*/
void
matrix3_orthographic_projection (double m[4][4],
				 double left, double right,
				 double bottom, double top,
				 double near, double far)
{
  /* pre */
  assert (m);
  assert (left < right);
  assert (bottom < top);
  assert (near < far);

  matrix3_initialize (m);

  m[0][0] = 2.0 / (right - left);
  m[3][0] = - (right + left) / (right - left);
  m[1][1] = 2.0 / (top - bottom);
  m[3][1] = - (top + bottom) / (top - bottom);
  m[2][2] = -2.0 / (far - near);
  m[3][2] = - (far + near) / (far - near);
}


/*------------------------------------------------------------*/
void
matrix3_transform (vector3 *v, double m[4][4])
{
  register double x, y, z, w;

  /* pre */
  assert (v);
  assert (m);

  x = v->x;
  y = v->y;
  z = v->z;
  w = x * m[0][3] + y * m[1][3] + z * m[2][3] + m[3][3];

  v->x = (x * m[0][0] + y * m[1][0] + z * m[2][0] + m[3][0]) / w;
  v->y = (x * m[0][1] + y * m[1][1] + z * m[2][1] + m[3][1]) / w;
  v->z = (x * m[0][2] + y * m[1][2] + z * m[2][2] + m[3][2]) / w;
}


/*------------------------------------------------------------*/
void
matrix3_rotate (vector3 *v, double m[4][4])
{
  register double x, y, z;

  /* pre */
  assert (v);
  assert (m);

  x = v->x;
  y = v->y;
  z = v->z;

  v->x = x * m[0][0] + y * m[1][0] + z * m[2][0];
  v->y = x * m[0][1] + y * m[1][1] + z * m[2][1];
  v->z = x * m[0][2] + y * m[1][2] + z * m[2][2];
}


/*------------------------------------------------------------*/
void
matrix3_rotate_translate (vector3 *v, double m[4][4])
{
  register double x, y, z;

  /* pre */
  assert (v);
  assert (m);

  x = v->x;
  y = v->y;
  z = v->z;

  v->x = x * m[0][0] + y * m[1][0] + z * m[2][0] + m[3][0];
  v->y = x * m[0][1] + y * m[1][1] + z * m[2][1] + m[3][1];
  v->z = x * m[0][2] + y * m[1][2] + z * m[2][2] + m[3][2];
}

/*------------------------------------------------------------*/
void
matrix3_invert (double dest[4][4], double src[4][4])
{
  double x[4][4], c[4][4], am, d;
  int ii, jj, i, j, i1, j1;

  for(ii = 0; ii<= 3; ii++) {
    for (jj = 0; jj<= 3; jj++) {
      i = 0;
      for (i1 = 0; i1 <= 3; i1++) {
      if(i1 != ii) {
        i++;
        j=0;
        for (j1 = 0; j1 <= 3; j1++) {
          if (j1 != jj) {
            j++;
            x[i-1][j-1] = src[i1][j1];
          }
        }
      }
      } /* 20 */
      am = x[0][0]*x[1][1]*x[2][2]-x[0][0]*x[1][2]*x[2][1]+
      x[0][1]*x[1][2]*x[2][0] - x[0][1]*x[1][0]*x[2][2]+
      x[0][2]*x[1][0]*x[2][1] - x[0][2]*x[1][1]*x[2][0];
      c[ii][jj] = ((ii + jj) % 2 == 0 ? 1 : -1) * am;
    } /* 30 */
  } /* 40 */

  /* Calculate determinant */
  d = 0;
  for (i=0; i<=3; i++) 
    d = d + src[i][0] * c[i][0];

  /* Invert the matrix */
  for (i=0; i<=3; i++) 
    for (j=0; j<=3; j++) 
      dest[i][j] = c[j][i]/d;
    
}

/*------------------------------------------------------------*/
void
matrix3_mult (double dest[4][4], double a[4][4], double b[4][4])
{
  int i, j, k;
  for(i=0 ; i<=3 ; i++)
    for(j=0 ; j<=3 ; j++) {
      dest[i][j] = 0;
      for(k=0 ; k<=3 ; k++) {
      dest[i][j] = dest[i][j] + b[i][k] * a[k][j];
      }
    }
}

/*------------------------------------------------------------*/
void
matrix2_invert (double dest[3][3], double src[3][3])
{
  int a, b, c, i, j;
  int size = 3;
  double **ident, **augment, **result;

  /*
   * TimF
   * I'll eventually rehash the matrix calls so they'll work for any
   * size matrix, so this has some extra bulkiness that will eventually
   * be reformatted as part of a library of memory alloc/dealloc funcs
   */
  if ((ident = calloc(size+1, sizeof(double *))) == NULL){
    perror("calloc");
    exit(1);
  }
  for (i=0; i<size; i++){
    if ((ident[i] = calloc(size+1, sizeof(double))) == NULL){
      perror("calloc");
      exit(1);
    }
  }

  if ((augment = calloc(size+1, sizeof(double *))) == NULL){
    perror("calloc");
    exit(1);
  }
  for (i=0; i<size; i++){
    if ((augment[i] = calloc(size*2+1, sizeof(double))) == NULL){
      perror("calloc");
      exit(1);
    }
  }

  if ((result = calloc(size+1, sizeof(double *))) == NULL){
    perror("calloc");
    exit(1);
  }
  for (i=0; i<size; i++){
    if ((result[i] = calloc(size*2+1, sizeof(double))) == NULL){
      perror("calloc");
      exit(1);
    }
  }

  /* let's do some multivariate analysis... */
  for (i=0; i<size; i++){
    for (j=0; j<size; j++){
      augment[i][j] = src[i][j];
      /* augment matrix with I */
      if (i == j)
	augment[i][j+size] = 1.0;
      else
	augment[i][j+size] = 0.0;
    }
  }

  for (j=0; j<size; j++){
    /* generate "new" identity */
    for (a=0; a<size; a++){
      for (b=0; b<size; b++){
	if (a == b)
	  ident[a][b] = 1.0;
	else
	  ident[a][b] = 0.0;
      }
    }

    /* generate (non-)pivot elements */
    for (i=0; i<size; i++){
      if (i == j)
	ident[i][j] = 1.0 / augment[i][j];
      else
	ident[i][j] = (-1.0 * augment[i][j]) / augment[j][j];
    }

    /* clear room for result */
    for (a=0; a<size; a++){
      for (b=0; b<size*2; b++)
	result[a][b] = 0.0;
    }

    /* multiply through (this could easily be a func) */
    for (c=0; c<size; c++){
      for (a=0; a<size*2; a++){
	for (b=0; b<size; b++){
	  result[c][a] += ident[c][b] * augment[b][a];
	}
      }
    }

    /* copy result to augmented matrix, and continue */
    for (a=0; a<size; a++){
      for (b=0; b<size*2; b++)
	augment[a][b] = result[a][b];
    }

  }

  /* return result */
  for (i=0; i<size; i++){
    for (j=size; j<size*2; j++)
      dest[i][j-size] = augment[i][j];
  }

  /* free up matrices (again, this will be part of a function call set sometime...) */
  for (i=0; i<size; i++)
    free(ident[i]);
  free(ident);
  for (i=0; i<size; i++)
    free(augment[i]);
  free(augment);
  for (i=0; i<size; i++)
    free(result[i]);
  free(result);
}

/*------------------------------------------------------------*/
void
matrix2_diagonalize (double evect[3][3], double A[3][3])
{
  int i, j;
  int niter = 0, size = 3;
  long max_iterations = 1000;
  double tolerance = 0.00001;
  double max_off_diag();

  /* generate I */
  for (i=0; i<size; i++){
    for (j=0; j<size; j++){
      if (i == j)
	evect[i][j] = 1.0;
      else
	evect[i][j] = 0.0;
    }
  }

  /* now diagonalize using jacobi rotation */
  while (niter++ < max_iterations && max_off_diag(A, &i, &j) > tolerance)
    jacobi_rotate (A, evect, i, j);
}

/*------------------------------------------------------------*/
double
max_off_diag (double A[3][3], int *maxi, int *maxj)
{
  int i, j;
  int size = 3;
  double max = 0.0;

  for (i=0; i<size; i++){
    for (j=i+1; j<size; j++){
      if (fabs(A[i][j]) > max){
	max = fabs(A[i][j]);
	*maxi = i;
	*maxj = j;
      }
    }
  }

  return (max);
}

/*------------------------------------------------------------*/
void
jacobi_rotate (double A[3][3], double X[3][3], int i, int j)
{
   int k;
   int size = 3;
   double theta, S, C, Xki, Aii;

   theta = 0.5 * atan(2 * A[i][j] / (A[i][i] - A[j][j]));
   S = sin(theta);
   C = cos(theta);

   for (k=0; k<size; k++){
     Xki = X[k][i];
     X[k][i] =   Xki * C + X[k][j] * S;
     X[k][j] = - Xki * S + X[k][j] * C;

     if (k != i && k != j){
       A[k][i] =   A[i][k] * C + A[j][k] * S;
       A[k][j] = - A[i][k] * S + A[j][k] * C;
       A[j][k] = A[k][j];
       A[i][k] = A[k][i];
     }
   }

   Aii = A[i][i];
   A[i][i] = Aii * C * C + A[j][j] * S * S + 2.0 * A[i][j] * S * C;
   A[j][j] = Aii * S * S + A[j][j] * C * C - 2.0 * A[i][j] * S * C;
   A[i][j] = A[j][i] = 0;
}
