/* delaunay.c
 *
 * delaunay tesselation of an atom selection
 * (requires GTS)
 *
 * 14-Mar-2003 TimF - first attempts
 */

#include <assert.h>
#include <stdlib.h>

#include "clib/extent3d.h"
#include "clib/matrix3.h"

#include "delaunay.h"
#include "global.h"
#include "graphics.h"
#include "lex.h"
#include "segment.h"
#include "select.h"
#include "state.h"
#include "xform.h"

#include <gts.h>

void delaunay(void){
  at3d **atoms;
  int atom_count, slot;
  int steiner_max = -1;
  guint num;
  gpointer data[2];

  gdouble quality = 0.0, area = G_MAXDOUBLE;

  GtsPoint *p;
  GtsSurface *surface;
  GtsTriangle *t;
  GtsVertex *tmp, *v, *v1, *v2, *v3;
  GSList *list = NULL;

  assert (count_atom_selections() == 1);

  atoms = select_atom_list (&atom_count);

  assert (count_atom_selections() == 0);

  if (message_mode)
    fprintf (stderr, "%i atoms selected for delaunay tesselation\n", atom_count);
  if (atoms == NULL) return;

  if (current_state->maptype == 0)
    line_segment_init();
  else
    output_start_object();

  /* 1: create triangle enclosing all verts */
  for (slot=0; slot<atom_count; slot++){
    p = gts_point_new(gts_point_class(), atoms[slot]->xyz.x, atoms[slot]->xyz.y, atoms[slot]->xyz.z);
    list = g_slist_prepend(list, p);
  }
  t = gts_triangle_enclosing(gts_triangle_class(), list, 100.0);
  g_slist_free(list);
  gts_triangle_vertices(t, &v1, &v2, &v3);

  /* create surface with one face: the enclosing triangle */
  surface = gts_surface_new(gts_surface_class(), gts_face_class(), gts_edge_class(), gts_vertex_class());
  gts_surface_add_face(surface, gts_face_new(gts_face_class(), t->e1, t->e2, t->e3));

  gts_surface_print_stats(surface, stderr);

  for (slot=0; slot<atom_count; slot++){
    v = gts_vertex_new(gts_vertex_class(), atoms[slot]->xyz.x, atoms[slot]->xyz.y, atoms[slot]->xyz.z);
    tmp = gts_delaunay_add_vertex(surface, v, NULL);

    g_assert (tmp != v);
    if (tmp != NULL){
      fprintf(stderr, "delaunay: duplicate vertex\n");
      gts_vertex_replace(v, tmp);
    }
  }

  if (gts_delaunay_check(surface)){
    fprintf (stderr, "delaunay: triangulation is not Delaunay\n");
    return;
  }

  gts_allow_floating_vertices = TRUE;
  gts_object_destroy(GTS_OBJECT(v1));
  gts_object_destroy(GTS_OBJECT(v2));
  gts_object_destroy(GTS_OBJECT(v3));
  gts_allow_floating_vertices = FALSE;

  num = gts_delaunay_conform(surface, steiner_max, (GtsEncroachFunc) gts_vertex_encroaches_edge, NULL);

  fprintf(stderr, "encroach: %d\n", num);

  if (num == 0){
    data[0] = &quality;
    data[1] = &area;
    num = gts_delaunay_refine(surface, steiner_max, (GtsEncroachFunc) gts_vertex_encroaches_edge, NULL, (GtsKeyFunc) triangle_cost, data);
  }
  gts_surface_print_stats(surface, stderr);

  data[0] = surface;
  gts_surface_foreach_face(surface, (GtsFunc) plot_delaunay, data);

  if (current_state->maptype == 0)
    output_line (FALSE);
  else
    output_finish_object();

  free (atoms);
}

void plot_delaunay(GtsTriangle *t, gpointer *data){
  int i;
  double x, y, z;
  vector3 object[9], tmp[6], u1, u2;
  colour mapcolour;
  line_segment *ls;
  GSList *j;
  GtsSurface *s = data[0];
  GtsVector norm;
  GtsVertex *v[3];

  gts_triangle_vertices(t, &v[0], &v[1], &v[2]);

  for (i=0; i<3; i++){
    norm[0] = norm[1] = norm[2] = 0.0;
    for (j=gts_vertex_faces(v[i], s, NULL); j; j=j->next){
      gts_triangle_normal(GTS_TRIANGLE((GtsFace *)j->data), &x, &y, &z);
      norm[0] += x;
      norm[1] += y;
      norm[2] += z;
    }

    u1.x = v[i]->p.x;
    u1.y = v[i]->p.y;
    u1.z = v[i]->p.z;
    u2.x = norm[0];
    u2.y = norm[1];
    u2.z = norm[2];

    object[i*3] = u1;

    colour_copy_to_rgb(&mapcolour, &current_state->linecolour);
    object[(i*3)+2].x = mapcolour.x;
    object[(i*3)+2].y = mapcolour.y;
    object[(i*3)+2].z = mapcolour.z;

    v3_normalize(&u2);
    object[(i*3)+1] = u2;
  }

  if (current_state->maptype == 0){
    ls = line_segment_next();
    ls->c = current_state->linecolour;
    ls->p = object[0];
    ls->new = 1;
    ext3d_update(&(ls->p), 0.0);
    ls = line_segment_next();
    ls->p = object[3];
    ext3d_update(&(ls->p), 0.0);
                                                                                
    ls = line_segment_next();
    ls->c = current_state->linecolour;
    ls->p = object[0];
    ext3d_update(&(ls->p), 0.0);
    ls = line_segment_next();
    ls->p = object[6];
    ext3d_update(&(ls->p), 0.0);
                                                                                
    ls = line_segment_next();
    ls->c = current_state->linecolour;
    ls->p = object[3];
    ext3d_update(&(ls->p), 0.0);
    ls = line_segment_next();
    ls->p = object[6];
    ext3d_update(&(ls->p), 0.0);
  }
  else{
    ext3d_update(&object[0], 0.0);
    ext3d_update(&object[3], 0.0);
    ext3d_update(&object[6], 0.0);

    if (current_state->smoothmap)
      output_object(OBJ_TRIANGLES_NORMALS_COLOURS, object, 9);
    else{
      /* in case we don't want normals */
      tmp[0] = object[0];
      tmp[2] = object[3];
      tmp[4] = object[6];
 
      tmp[1] = object[2];
      tmp[3] = object[5];
      tmp[5] = object[8];
 
      output_object(OBJ_TRIANGLES_COLOURS, tmp, 6);
    }
  }
}

gdouble triangle_cost(GtsTriangle *t, gpointer *data){
  gdouble *min_quality = data[0];
  gdouble *max_area = data[1];
  gdouble quality = gts_triangle_quality(t);
  gdouble area = gts_triangle_area(t);

  if (quality < *min_quality || area > *max_area)
    return quality;

  return 0.0;
}
