/* edmap.c
 *
 * read in:
 * DSN6 (a.k.a. BRICK)
 * BRIX
 * OpenDX scalar field files
 *
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gts.h>

#include "clib/aa_lookup.h"
#include "clib/str_utils.h"
#include "clib/extent3d.h"
#include "clib/matrix3.h"
#include "clib/mol3d.h"

#include "edmap.h"
#include "global.h"
#include "graphics.h"
#include "map_utils.h"
#include "segment.h"
#include "select.h"
#include "state.h"
#include "vertex_class.h"

/* so math.h functions are called correctly */
extern double round(double x);

FILE *ifp;
int totr=0;

/* compute cell volume, setup tranformation matrices for scalar maps that require it (e.g. e.d. maps) */
void compute_orthog(void){
   double ca[3], sa[3], cas[3], vol, sas0;
   int i, j;

   /* create cos and sin table */
   for (i=0; i<3; i++) {
     ca[i] = cos(curmap->cangles[i] * TORAD);
     sa[i] = sin(curmap->cangles[i] * TORAD);
   }
 
   /* create cos(angle) */
   cas[0] = (ca[1] * ca[2] - ca[0]) / (sa[1] * sa[2]);
   cas[1] = (ca[2] * ca[0] - ca[1]) / (sa[2] * sa[0]);
   cas[2] = (ca[0] * ca[1] - ca[2]) / (sa[0] * sa[1]);
 
   /* compute cell volume */
   vol = curmap->clen[0] * curmap->clen[1] * curmap->clen[2] * sqrt(1.0 + 2.0 * ca[0] * ca[1] * ca[2] - ca[0] * ca[0] - ca[1] * ca[1] - ca[2] * ca[2]);
   sas0 = sqrt(1.0 - cas[0] * cas[0]);

   #ifdef EDEBUG
   fprintf(stderr, "Volume is %f\n", vol);
   fprintf(stderr, "ca is %f %f %f\n", ca[0], ca[1], ca[2]);
   fprintf(stderr, "sa is %f %f %f\n", sa[0], sa[1], sa[2]);
   fprintf(stderr, "cas is %f %f %f\n", cas[0], cas[1], cas[2]);
   #endif

   /* set up matrices */
   curmap->f_t_c[0][0] = curmap->clen[0];
   curmap->f_t_c[1][0] = ca[2] * curmap->clen[1];
   curmap->f_t_c[2][0] = ca[1] * curmap->clen[2];
   curmap->f_t_c[1][1] = sa[2] * curmap->clen[1];
   curmap->f_t_c[2][1] = -sa[1] * cas[0] * curmap->clen[2];
   curmap->f_t_c[2][2] = sa[1] * sas0 * curmap->clen[2];

   curmap->c_t_f[0][0] = 1.0 / curmap->clen[0];
   curmap->c_t_f[1][0] = -ca[2] / (sa[2] * curmap->clen[0]);
   curmap->c_t_f[2][0] = -(ca[2] * sa[1] * cas[0] + ca[1] * sa[2]) / (sa[1] * sas0 * sa[2] * curmap->clen[0]);
   curmap->c_t_f[1][1] = 1.0 / (sa[2] * curmap->clen[1]);
   curmap->c_t_f[2][1] = cas[0] / (sas0 * sa[2] * curmap->clen[1]);
   curmap->c_t_f[2][2] = 1.0 / (sa[1] * sas0 * curmap->clen[2]);

   #ifdef EDEBUG
   fprintf(stderr, "f_t_c:\n");
   for (i=0; i<3; i++){
     for (j=0; j<3; j++){
       fprintf(stderr, "%f ", curmap->f_t_c[i][j]);
     }
     fprintf(stderr, "\n");
   }

   fprintf(stderr, "c_t_f:\n");
   for (i=0; i<3; i++){
     for (j=0; j<3; j++){
       fprintf(stderr, "%f ", curmap->c_t_f[i][j]);
     }
     fprintf(stderr, "\n");
   }
   #endif 
}

/* read in map */
void open_map(char *filename){
  int sel, atom_count, slot;
  vector3 low, high;
  at3d **atoms;

  sel = count_atom_selections();
  atom_count = 0;

  /* molscript initialization */
  if (sel){
    assert (sel == 1);

    ext3d_push();
    ext3d_initialize();

    atoms = select_atom_list (&atom_count);
    if (message_mode)
      fprintf (stderr, "%i atoms selected for map reading\n", atom_count);
 
    for (slot = 0; slot < atom_count; slot++)
      ext3d_update (&(atoms[slot]->xyz), plus_level);

    free(atoms);

    ext3d_get_low_high(&low, &high);
    ext3d_pop (FALSE);
    assert (count_atom_selections() == 0);
  }

  if ((ifp = gfopen(filename, "rb")) == NULL){
    free_data(curmap->name);
    fprintf(stderr, "e- density map file %s does not exist or is not readable\n", filename);
    exit_on_error = TRUE;
    yyerror("e- density file invalid");
  }

  /* read and close the file */
  if (sel && atom_count){
    fprintf(stderr, "read %s with low and high limits\n", filename);
    read_header(&low, &high);
  }
  else{
    fprintf(stderr, "read %s with no limits\n", filename);
    read_header(NULL, NULL);
  }

  if (curmap->format == APBS)
    read_dx_data();
  else
    read_edmap_data();
  fclose(ifp);
}

/* read in OpenDX file format scalar field */
void read_dx_data(void){
  int i, j, k, n;
  char buf[MAXLINE];
  double val[3];

  /* initialize 3D array */
  if ((curmap->data = calloc(curmap->len[0], sizeof(double **))) == NULL){
    perror("calloc");
    exit(1);
  }
  for (i=0; i<curmap->len[0]; i++){
    if ((curmap->data[i] = calloc(curmap->len[1], sizeof(double *))) == NULL){
      perror("calloc");
      exit(1);
    }
    for (j=0; j<curmap->len[1]; j++){
      if ((curmap->data[i][j] = calloc(curmap->len[2], sizeof(double))) == NULL){
        perror("calloc");
        exit(1);
      }

      /* init k's to 0 */
      for (k=0; k<curmap->len[2]; k++)
        curmap->data[i][j][k] = 0.0;
    }
  }

  /* read in data */
  i = j = k = 0;
  while (i < curmap->len[0]){
    if ((fgets(buf, MAXLINE, ifp)) == NULL){
      fprintf(stderr, "%d %d %d  %d %d %d\n", i, j, k, curmap->len[0], curmap->len[1], curmap->len[2]);
      fprintf(stderr, "OpenDX file ended prematurely\n");
      return;
    }

    /* 3 points per line */
    sscanf(buf, "%le %le %le", &val[0], &val[1], &val[2]);
    for (n=0; n<3; n++){
      curmap->data[i][j][k] = val[n];

      if (k+1 < curmap->len[2])
	k++;
      else{
	k=0;
	if (j+1 < curmap->len[1])
	  j++;
	else{
	  j=0;
	  i++;
	  if (i >= curmap->len[0])
	    break;
	}
      }
    }
  }
}

/* read in an *entire* electron density map */
void read_edmap_data(void){
  int i, j, k;
  int x, y, z;
  int curx, cury, curz;
  int n;
  int xext, yext, zext;
  int xnum, ynum, znum;
  int bytes_read=0;
  unsigned char buf[BRICK];
  boolean toswap=TRUE;

  /* set extent of map and number of records within that extent */
  xext=curmap->len[0];
  xnum=((xext) / BSIZE) + (xext % 8 ? 1 : 0);
  yext=curmap->len[1];
  ynum=((yext) / BSIZE) + (yext % 8 ? 1 : 0);
  zext=curmap->len[2];
  znum=((zext) / BSIZE) + (zext % 8 ? 1 : 0);

  if ((xext<0) || (yext<0) || (zext<0) || (xnum<0) || (ynum<0) || (znum<0)){
    fprintf(stderr, "error in settings - region to contour outside of map\n");
    return;
  }

  /* allocate space for the map as a 3D pointer array */
  if ((curmap->data = calloc(xext+1, sizeof(double **))) == NULL){
    perror("calloc");
    exit(1);
  }
  for (i=0; i<xext+1; i++){
    if ((curmap->data[i] = calloc(yext+1, sizeof(double *))) == NULL){
      perror("calloc");
      exit(1);
    }
    for (j=0; j<yext+1; j++){
      if ((curmap->data[i][j] = calloc(zext+1, sizeof(double))) == NULL){
	perror("calloc");
	exit(1);
      }
      /* initialize all k to zero here if desired */
    }
  }

  #ifdef EDEBUG
  fprintf(stderr, "xext: %d xnum: %d\nyext: %d ynum: %d\nzext: %d znum: %d\n\n", xext, xnum, yext, ynum, zext, znum);
  #endif

  /* for all bricks */
  for (z=0; z<znum; z++){
    /* current Z in terms of bricks */
    curz=BSIZE*z;
    for (y=0; y<ynum; y++){
      /* current Y in terms of bricks */
      cury=BSIZE*y;
      for (x=0; x<xnum; x++){
	/* current X in terms of bricks */
	curx=BSIZE*x;

	/* read in brick at curx,cury,curz */
	bytes_read = fread(buf, sizeof(char), BRICK, ifp);
	totr += bytes_read;

	/* buffer index */
	n=0;
	/* interpret the current brick */
	for (i=0; i<BSIZE; i++){
	  for (j=0; j<BSIZE; j++){
	    for (k=0; k<BSIZE; k++){
	      if (((curz+i)>zext) || ((cury+j)>yext) || ((curx+k)>xext))
		/* out of map bounds */
		;
	      else{
		if (curmap->format == DSN6){
		  /*
		   * the DSN6 format is weird...
		   * bytes are swapped, even though the data is stored in single bytes
		   * (i.e. technically, they *can't* be swapped)
		   * therefore, we need to swap bytes *across the array*...
		   *
		   * argh.
		   */
		  if (toswap)
		    swap((char *)&buf[n], 2);
		  toswap = (toswap ? FALSE : TRUE);
		}
		curmap->data[curx+k][cury+j][curz+i]=(buf[n] - curmap->plus) / curmap->prod;
	      }
	      n++;
	    }
	  }
	}

      }
    }
  }
  #ifdef EDEBUG
  fprintf(stderr, "bytes read: %d tot read: %d\n\n", bytes_read, totr);
  #endif
}

/* read in scalar field header */
void read_header(vector3 *low, vector3 *high){
  int i;
  int bytes_read;
  int origin[3], sampling[3], gridpos[3];
  unsigned short hdr[BRICK];
  char buf[MAXLINE];
  char *maphdr;

  double val1, val2;

  /* read in full header */
  bytes_read = fread((char *)&hdr, sizeof(short), BRICK/2, ifp);
  totr += bytes_read;

  if ((strstr((char *) &hdr[0], ":-)")) != NULL){
    #ifdef EDEBUG
    fprintf(stderr, "%s is BRIX map format\n", curmap->name);
    #endif
    curmap->format=BRIX;
  }
  else if ((strstr((char *) &hdr[0], "# Data from APBS")) != NULL){
    #ifdef EDEBUG
    fprintf(stderr, "%s is OpenDX (APBS) scalar format\n", curmap->name);
    #endif
    rewind(ifp);
    curmap->format=APBS;
  }
  else{
    #ifdef EDEBUG
    fprintf(stderr, "%s is DSN6 map format\n", curmap->name);
    #endif
    curmap->format=DSN6;
  }

  /*
   * swap bytes (if necessary)
   * assumes all DSN6 input files are big endian
   * (theoretically true for all DSN6 files made in the "O" days)
   */
  #ifndef WORDS_BIGENDIAN
  if (curmap->format == DSN6){
    for (i=0; i<19; i++)
      swap((char *)&hdr[i], sizeof(short));
  }
  #endif

  /* read in the header */
  if (curmap->format == BRIX){
    strtok((char *) &hdr[0], " ");
    for (i=0; ((maphdr = strtok(NULL, " ")) != NULL); i++){
      if ((i>0) && (i<4))
	origin[i-1]=atoi(maphdr);
      else if ((i>4) && (i<8))
	curmap->len[i-5]=atoi(maphdr);
      else if ((i>8) && (i<12))
	sampling[i-9]=atoi(maphdr);
      else if ((i>12) && (i<16))
	curmap->clen[i-13]=atof(maphdr);
      else if ((i>15) && (i<19))
	curmap->cangles[i-16]=atof(maphdr);
      else if (i == 20)
	curmap->prod=atof(maphdr);
      else if (i == 22)
	curmap->plus=(double) atoi(maphdr);
      else if (i == 24)
	curmap->sigma=atof(maphdr);
    }

    for (i=0; i<3; i++){
      curmap->ori[i] = (double) origin[i] / (double) sampling[i];
      curmap->diff[i] = 1.0 / (double) sampling[i];
    }
  }
  else if (curmap->format == DSN6){
    #ifdef EDEBUG
    for (i=0; i<19; i++)
      fprintf(stderr, "hdr(%d): %d\n", i+1, hdr[i]);
    #endif

    /* setup cell and prod value */
    for (i=0; i<3; i++)
      curmap->clen[i] = (double) hdr[i+9] / (double) hdr[17];
    for (i=3; i<6; i++)
      curmap->cangles[i-3] = (double) hdr[i+9] / (double) hdr[17];

    curmap->prod = (double) hdr[15] / (double) hdr[18];
    curmap->plus = (double) hdr[16];

    /*
     * watch out for integer overflows (2 byte word machines)
     * store new origin in a signed int value
     */
    for (i=0; i<3; i++){
      if (hdr[i] > 32768)
	origin[i] = hdr[i] - 65536;
      else
	origin[i] = hdr[i];
      curmap->len[i] = hdr[i+3];
    }

    for (i=0; i<3; i++){
      curmap->ori[i] = (double) origin[i] / (double) hdr[i+6];
      curmap->diff[i] = 1.0 / (double) hdr[i+6];
    }
  }
  else if (curmap->format == APBS){
    /* keep reading in header until data is about to appear */
    while (1){
      if ((fgets(buf, MAXLINE, ifp)) == NULL){
	fprintf(stderr, "OpenDX file ended prematurely\n");
	return;
      }

      if ((strstr((char *) &buf, "object 1 class gridpositions counts")) != NULL)
	sscanf(buf, "object 1 class gridpositions counts %d %d %d", &curmap->len[0], &curmap->len[1], &curmap->len[2]);

      if ((strstr((char *) &buf, "origin")) != NULL)
	sscanf(buf, "origin %le %le %le", &curmap->ori[0], &curmap->ori[1], &curmap->ori[2]);

      if ((strstr((char *) &buf, "delta")) != NULL){
	sscanf(buf, "delta %le %le %le", &curmap->diff[0], &val1, &val2);
	fgets(buf, MAXLINE, ifp);
	sscanf(buf, "delta %le %le %le", &val1, &curmap->diff[1], &val2);
	fgets(buf, MAXLINE, ifp);
	sscanf(buf, "delta %le %le %le", &val1, &val2, &curmap->diff[2]);
      }

      if ((strstr((char *) &buf, "data follows")) != NULL){
	break;
      }
    }
  }

  #ifdef EDEBUG
  for (i=0; i<3; i++)
    fprintf(stderr, "origin: %f\n", curmap->ori[i]);
  for (i=0; i<3; i++)
    fprintf(stderr, "extent: %d\n", curmap->len[i]);
  for (i=0; i<3; i++)
    fprintf(stderr, "sampling: %f\n", curmap->diff[i]);

  fprintf(stderr, "\na: %f b: %f c: %f \n", curmap->clen[0], curmap->clen[1], curmap->clen[2]);
  fprintf(stderr, "alpha: %f beta: %f gamma: %f \n\n", curmap->cangles[0], curmap->cangles[1], curmap->cangles[2]);
  #endif

  /* compute cell matrices */
  if (curmap->format == BRIX || curmap->format == DSN6){
    fprintf(stderr, "prod: %f plus: %f\n", curmap->prod, curmap->plus);
    compute_orthog();
  }

  if (low){
    /*
     * if there was a selection:
     * 1) apply appropriate transformations
     * 2) set up transformed headers
     */
    matrix3_transform(low, curmap->c_t_f);
    matrix3_transform(high, curmap->c_t_f);

    /* new origin */
    gridpos[0] = (int) ((low->x - curmap->ori[0]) / curmap->diff[0]);
    gridpos[1] = (int) ((low->y - curmap->ori[1]) / curmap->diff[1]);
    gridpos[2] = (int) ((low->z - curmap->ori[2]) / curmap->diff[2]);

    for (i=0; i<3; i++){
      if (gridpos[i] >= 0 && gridpos[i] < curmap->len[i])
	curmap->offset[i] = gridpos[i];
      else
	curmap->offset[i] = 0;
    }

    /* new extent */
    gridpos[0] = (int) ((high->x - curmap->ori[0]) / curmap->diff[0]);
    gridpos[1] = (int) ((high->y - curmap->ori[1]) / curmap->diff[1]);
    gridpos[2] = (int) ((high->z - curmap->ori[2]) / curmap->diff[2]);

    for (i=0; i<3; i++){
      if (gridpos[i] > 0 && gridpos[i] <= curmap->len[i] && gridpos[i] > curmap->offset[i])
	curmap->offset[i+3] = gridpos[i];
      else
	curmap->offset[i+3] = curmap->len[i];
    }
  }
  else{
    for (i=0; i<3; i++){
      curmap->offset[i] = 0;
      curmap->offset[i+3] = curmap->len[i];
    }
  }

  #ifdef EDEBUG
  for (i=0; i<6; i++)
    fprintf(stderr, "offset %d: %d\n", i+1, curmap->offset[i]);
  #endif
}
