/* gensurf.c
 *
 * generate a surface from an atom selection
 *
 * 12-Apr-2004 TimF - first attempts
 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gts.h>
 
#include "clib/extent3d.h"
#include "clib/mol3d.h"
#include "clib/vector3.h"

#include "gensurf.h"
#include "global.h"
#include "graphics.h"
#include "map_utils.h"
#include "segment.h"
#include "select.h"
#include "state.h"
#include "vertex_class.h"

/* so math.h functions are called correctly */
extern double round(double x);

/* allocate 3D data for generated surface map */
void gensurf_maplen(float rate, vector3 *min, vector3 *max){
  int i, j, k;
  double inv, len[3];

  curmap->format = MOLSURF;

  /* set up map extent */
  #ifdef EDEBUG
  fprintf(stderr, "surface map:\n");
  fprintf(stderr, "rate: %3.2f range:\n", rate);
  fprintf(stderr, "x: %f %f\n", min->x, max->x);
  fprintf(stderr, "y: %f %f\n", min->y, max->y);
  fprintf(stderr, "z: %f %f\n", min->z, max->z);
  #endif

  len[0] = round((max->x - min->x) * rate);
  len[1] = round((max->y - min->y) * rate);
  len[2] = round((max->z - min->z) * rate);

  /* save information in structure */
  for (i=0; i<3; i++){
    curmap->len[i] = (int) len[i];
    curmap->diff[i] = 1.0 / ((double) rate);
  }
  curmap->ori[0] = min->x;
  curmap->ori[1] = min->y;
  curmap->ori[2] = min->z;
  curmap->offset[0] = 0;
  curmap->offset[1] = 0;
  curmap->offset[2] = 0;
  curmap->offset[3] = curmap->len[0];
  curmap->offset[4] = curmap->len[1];
  curmap->offset[5] = curmap->len[2];

  #ifdef EDEBUG
  fprintf(stderr, "len: %d %d %d\n", (int) (len[0]), (int) (len[1]), (int) (len[2]));
  #endif

  /* initialize 3D array */
  if ((curmap->data = calloc(((int) len[0]), sizeof(double **))) == NULL){
    perror("calloc");
    exit(1);
  }
  for (i=0; i<((int)len[0]); i++){
    if ((curmap->data[i] = calloc(((int) len[1]), sizeof(double *))) == NULL){
      perror("calloc");
      exit(1);
    }
    for (j=0; j<((int)len[1]); j++){
      if ((curmap->data[i][j] = calloc(((int) len[2]), sizeof(double))) == NULL){
	perror("calloc");
	exit(1);
      }

      /* init k's to 0 */
      for (k=0; k<((int) len[2]); k++)
	curmap->data[i][j][k] = 0.0;
    }
  }
}


/* generate map */
void gensurf_init(void){
  int i, j, k, x1, x2, y1, y2, z1, z2;
  int sel, atom_count, slot;
  double val;
  vector3 low, high, atompt, mappt;
  at3d **atoms;

  sel = count_atom_selections();
  atom_count = 0;

  /* molscript initialization */
  assert (sel == 1);

  ext3d_push();
  ext3d_initialize();

  atoms = select_atom_list (&atom_count);
  if (message_mode)
    fprintf (stderr, "%i atoms selected for surface generation (sampling: %3.2f)\n", atom_count, current_state->sampling);

  for (slot = 0; slot < atom_count; slot++)
    ext3d_update (&(atoms[slot]->xyz), plus_level);

  ext3d_get_low_high(&low, &high);
  ext3d_pop (FALSE);
  assert (count_atom_selections() == 0);

  /* make the map structure */
  gensurf_maplen(1.0/current_state->sampling, &low, &high);

  for (slot = 0; slot < atom_count; slot++){
    atompt.x = atoms[slot]->xyz.x;
    atompt.y = 0.0;
    atompt.z = 0.0;

    mappt.x = curmap->ori[0];
    mappt.y = 0.0;
    mappt.z = 0.0;

    /* find map points near the atom */
    for (x1=0; (x1<curmap->len[0]) && (v3_distance(&atompt, &mappt) > atoms[slot]->radius+current_state->probe_rad+0.5); x1++)
      mappt.x += curmap->diff[0];
    for (x2=x1; (x2<curmap->len[0]) && !(v3_distance(&atompt, &mappt) > atoms[slot]->radius+current_state->probe_rad+0.5); x2++)
      mappt.x += curmap->diff[0];
    mappt.x = curmap->ori[0] + (curmap->diff[0]*(((double) x1 + x2)/2.0));

    atompt.y = atoms[slot]->xyz.y;
    mappt.y = curmap->ori[1];

    for (y1=0; (y1<curmap->len[1]) && (v3_distance(&atompt, &mappt) > atoms[slot]->radius+current_state->probe_rad+0.5); y1++)
      mappt.y += curmap->diff[1];
    for (y2=y1; (y2<curmap->len[1]) && !(v3_distance(&atompt, &mappt) > atoms[slot]->radius+current_state->probe_rad+0.5); y2++)
      mappt.y += curmap->diff[1];
    mappt.y = curmap->ori[1] + (curmap->diff[1]*(((double) y1 + y2)/2.0));

    atompt.z = atoms[slot]->xyz.z;
    mappt.z = curmap->ori[2];

    for (z1=0; (z1<curmap->len[2]) && (v3_distance(&atompt, &mappt) > atoms[slot]->radius+current_state->probe_rad+0.5); z1++)
      mappt.z += curmap->diff[2];
    for (z2=z1; (z2<curmap->len[2]) && !(v3_distance(&atompt, &mappt) > atoms[slot]->radius+current_state->probe_rad+0.5); z2++)
      mappt.z += curmap->diff[2];

    /* now fill those map points */
    for (i=x1; i<x2; i++){
      mappt.x = curmap->ori[0] + (curmap->diff[0]*(double) i);
      for (j=y1; j<y2; j++){
	mappt.y = curmap->ori[1] + (curmap->diff[1]*(double) j);
	for (k=z1; k<z2; k++){
	  mappt.z = curmap->ori[2] + (curmap->diff[2]*(double) k);

	  /*
	   * equation to generate scalars:
	   * density = (1-(distance/radius)^2)^2
	   */
	  val = (v3_distance(&atompt, &mappt)/(current_state->probe_rad+atoms[slot]->radius) > 1.0) ? 0.0 :
	    pow(1.0 - pow(v3_distance(&atompt, &mappt)/(current_state->probe_rad+atoms[slot]->radius), 2.0), 2.0);
	  curmap->data[i][j][k] += val;
	}
      }
    }

  }
  free(atoms);
}
