/* map_utils.c
 *
 * map utilities (mostly for scalar fields and GTS surfaces)
 *
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gts.h>

#include "clib/aa_lookup.h"
#include "clib/extent3d.h"
#include "clib/matrix3.h"
#include "clib/str_utils.h"
#include "clib/vector3.h"

#include "global.h"
#include "graphics.h"
#include "map_utils.h"
#include "segment.h"
#include "select.h"
#include "state.h"

map_ptr map1, curmap;
float contour_level = 1.0;
double plus_level = 0.0;

/* add map to linked list data array */
void add_map (char *name){
  int i;

  /* initialize linked list or add an element to the list */
  fprintf(stderr, "allocating map space for %s\n", name);
  if (map1 == NULL){
    if ((curmap = malloc(sizeof(map))) == NULL){
      perror("malloc");
      exit(1);
    }
    map1 = curmap;
  }
  else{
    curmap = map1;
    while (curmap->next != NULL)
      curmap = curmap->next;
    if ((curmap->next = malloc(sizeof(map))) == NULL){
      perror("malloc");
      exit(1);
    }
    curmap = curmap->next;
  }
  curmap->name = str_clone(name);
  curmap->format = NONE;
  curmap->drawflag = 0;
  curmap->lambda = 0.0;
  curmap->coldef = NULL;
  for (i=0; i<3; i++){
    curmap->clen[i] = 0.0;
    curmap->cangles[i] = 0.0;
    curmap->offset[i] = 0;
    curmap->offset[i+3] = 0;
  }
  matrix3_initialize(curmap->c_t_f);
  matrix3_initialize(curmap->f_t_c);
  curmap->surface = NULL;
  curmap->data = NULL;
  curmap->next = NULL;
}

/* cost function based on angle (for surface simplification) */
double cost_angle (GtsEdge *e){
  if (e->triangles && e->triangles->next)
    return fabs (gts_triangles_angle (e->triangles->data,
                                      e->triangles->next->data));
  return G_MAXDOUBLE;
}

/* draw map */
void draw_map(char *id){
  int i, sel, slot, atom_count;
  int gridpos[3], offset[6];
  double edgenum;
  at3d **mapatoms = NULL;
  vector3 low, high, lowres, highres;

  GtsCartesianGrid g;

  GtsKeyFunc cost_func = (GtsKeyFunc) gts_volume_optimized_cost;
  /* GtsKeyFunc cost_func = (GtsKeyFunc) cost_angle; */
  GtsCoarsenFunc coarsen_func = (GtsCoarsenFunc) gts_volume_optimized_vertex;
  gpointer stop_data = NULL;

  GtsVolumeOptimizedParams params = {0.5, 0.5, 1e-10};
  gpointer cost_data = &params, coarsen_data = &params;
  /* gpointer cost_data = NULL, coarsen_data = &params; */

  /* molscript setup */
  sel = count_atom_selections();
  atom_count = 0;

  gpointer data[4];

  if (sel) {
    assert (sel == 1);
 
    ext3d_push();
    ext3d_initialize();
 
    mapatoms = select_atom_list (&atom_count);
 
    if (message_mode)
      fprintf (stderr, "%i atoms selected for map covering\n", atom_count);
 
    for (slot = 0; slot < atom_count; slot++)
      ext3d_update (&(mapatoms[slot]->xyz), current_state->mapradius);
 
    ext3d_get_low_high(&low, &high);
    ext3d_pop (FALSE);
    assert (count_atom_selections() == 0);
  }

  /* find the map */
  if (map1 == NULL){
    fprintf(stderr, "map %s not found in stack\n", id);
    return;
  }
  curmap = map1;
  while (strncmp(curmap->name, id, strlen(id)) != 0){
    if (curmap->next == NULL){
      fprintf(stderr, "map %s not found in stack\n", id);
      return;
    }
    curmap = curmap->next;
  }

  data[0] = &curmap;

  /* if its one of the following formats, need to polygonize */
  if (curmap->format == BRIX || curmap->format == DSN6 || curmap->format == APBS || curmap->format == MOLSURF){

    if (sel && atom_count){

      map_get_low_high(&low, &high, &lowres, &highres);

#ifdef EDEBUG
      fprintf(stderr, "low: %f %f %f\n", lowres.x, lowres.y, lowres.z);
      fprintf(stderr, "high: %f %f %f\n", highres.x, highres.y, highres.z);
#endif

      /* lowres - ori * 1/rate - find origin of current contour */
      gridpos[0] = (int) ((lowres.x - curmap->ori[0]) / curmap->diff[0]);
      gridpos[1] = (int) ((lowres.y - curmap->ori[1]) / curmap->diff[1]);
      gridpos[2] = (int) ((lowres.z - curmap->ori[2]) / curmap->diff[2]);

      for (i=0; i<3; i++){
	if (gridpos[i] >= curmap->offset[i] && gridpos[i] < curmap->offset[i+3])
	  offset[i] = gridpos[i];
	else
	  offset[i] = curmap->offset[i];
      }

      /* highres - ori * 1/rate - find extent of contour */
      gridpos[0] = (int) ((highres.x - curmap->ori[0]) / curmap->diff[6]);
      gridpos[1] = (int) ((highres.y - curmap->ori[1]) / curmap->diff[1]);
      gridpos[2] = (int) ((highres.z - curmap->ori[2]) / curmap->diff[2]);

      for (i=0; i<3; i++){
	if (gridpos[i] > curmap->offset[i] && gridpos[i] < curmap->offset[i+3])
	  offset[i+3] = gridpos[i];
	else
	  offset[i+3] = curmap->offset[i+3];

	offset[i+3] -= offset[i];
      }
    }
    else{
      for (i=0; i<6; i++)
	offset[i] = curmap->offset[i];

      for (i=0; i<3; i++)
	offset[i+3] -= offset[i];
    }

#ifdef EDEBUG
    fprintf(stderr, "contouring xorg: %d xext: %d yorg: %d yext: %d zorg: %d zext: %d\n",
	    offset[0], offset[3], offset[1], offset[4], offset[2], offset[5]);
#endif

    /* write density file for PovRay */
    if (current_state->maptype == 2){
      fprintf(stderr, "generating df3 output for %s\n", curmap->name);
      write_df3(&mapatoms, atom_count, offset[0], offset[0]+offset[3], offset[1], offset[1]+offset[4], offset[2], offset[2]+offset[5]);
      return;
    }

    /* setup for GTS */
    g.dx = curmap->diff[0];
    g.dy = curmap->diff[1];
    g.dz = curmap->diff[2];
    g.nx = offset[3];
    g.ny = offset[4];
    g.nz = offset[5];
    g.x = (double) (offset[0] * curmap->diff[0]) + curmap->ori[0];
    g.y = (double) (offset[1] * curmap->diff[1]) + curmap->ori[1];
    g.z = (double) (offset[2] * curmap->diff[2]) + curmap->ori[2];

    data[0] = &curmap;
    data[1] = &offset[0];
    data[2] = &offset[1];
    data[3] = &offset[2];

    curmap->surface = gts_surface_new(gts_surface_class(), gts_face_class(), gts_edge_class(), GTS_VERTEX_CLASS(gts_vertex2_class()));

    /* polygonize */
    contour_level = (curmap->format == MOLSURF) ? 0.1 : contour_level;
    if (current_state->tetra)
      gts_isosurface_tetra_bcl(curmap->surface, g, (GtsIsoCartesianFunc) get_map_plane, data, contour_level);
    else
      gts_isosurface_cartesian(curmap->surface, g, (GtsIsoCartesianFunc) get_map_plane, data, contour_level);

    gts_surface_foreach_face(curmap->surface, (GtsFunc) revert_tri, data);

    /* refine/simplify */
    fprintf(stderr, "refining surface: %s...\n", curmap->name);
    if (current_state->mapthresh > 0.0){
      edgenum = gts_surface_edge_number(curmap->surface) - (current_state->mapthresh * gts_surface_edge_number(curmap->surface));
      stop_data = &edgenum;
      gts_surface_coarsen(curmap->surface, cost_func, cost_data, coarsen_func, coarsen_data, (GtsStopFunc) stop_func, stop_data, M_PI/180.0);
    }
    else if (current_state->mapthresh < 0.0){
      edgenum = (current_state->mapthresh * gts_surface_edge_number(curmap->surface)) - gts_surface_edge_number(curmap->surface);
      stop_data = &edgenum;
      gts_surface_refine(curmap->surface, NULL, NULL, NULL, NULL, (GtsStopFunc) stop_func, stop_data);
    }
  }

  data[1] = &atom_count;
  data[2] = &mapatoms;
  /*
   * figure out which tris to draw and delete triangles if necessary
   */
  gts_surface_foreach_face(curmap->surface, (GtsFunc) set_drawable, data);

  if (curmap->format == BRIX || curmap->format == DSN6 || curmap->format == MOLSURF)
    gts_surface_foreach_face_remove(curmap->surface, (GtsFunc) remove_tri, data);

  if (curmap->lambda > 0.0){
    fprintf(stderr, "smoothing surface (i.e. Laplacian filter using lambda: %f)\n", curmap->lambda);
    data[0] = curmap->surface;
    gts_surface_foreach_vertex(curmap->surface, (GtsFunc) smooth_vertex, data);
  }

  fprintf(stderr, "setting vertex attributes...\n");
  if (curmap->drawflag == SCALARS)
    fprintf(stderr, "mapping scalars from %s onto %s\n", curmap->scalars->name, curmap->name);
  data[0] = &curmap;
  gts_surface_foreach_vertex(curmap->surface, (GtsFunc) set_vertex_attributes, data);

  fprintf(stderr, "surface statistics for %s:\n", curmap->name);
  gts_surface_print_stats(curmap->surface, stderr);
  fprintf(stderr, "is a closed surface: %s\n", gts_surface_is_closed(curmap->surface) ? "true" : "false");
  fprintf(stderr, "  (if false, volume value probably invalid)\n");
  fprintf(stderr, "surface area (A^2): %f\n", gts2_surface_area(curmap->surface));
  fprintf(stderr, "surface volume (A^3): %f\n", fabs(gts2_surface_volume(curmap->surface)));

  /* setup output */
  if ((curmap->format == GRASP) || (curmap->drawflag == CURVATURES)){
    fprintf(stderr, "drawing surface with colour scale %f low, %f mid and %f high\n(input surface levels: low: %f high %f)\n",
	    current_state->surflow, current_state->surfmid, current_state->surfhigh, curmap->collow, curmap->colhigh);
  }

  output_gtssurface(&curmap);

  /* cleanup */
  if ((curmap->format == BRIX || curmap->format == DSN6 || curmap->format == MOLSURF) && curmap->surface){
    gts_object_destroy(GTS_OBJECT(curmap->surface));
    curmap->surface=NULL;
  }
  if (sel){
    free(mapatoms);
    mapatoms = NULL;
  }
}

/* free all maps in the stack */
void free_map_all(void){
  #ifdef EDEBUG
  fprintf(stderr, "map stack: ");
  for (curmap=map1; curmap != NULL; curmap=curmap->next)
    fprintf(stderr, "%s->", curmap->name);
  fprintf(stderr, "NULL\n");
  #endif

  for (curmap=map1; curmap != NULL;)
    free_data(curmap->name);
}

/* free map data */
void free_data(char *id){
  int add, i, j;
  map_ptr tmp;

  /* find the map */
  curmap = map1;

  /* any maps in the stack? */
  if (curmap == NULL){
    fprintf(stderr, "map %s init not found in stack\n", id);
    return;
  }

  while (strncmp(curmap->name, id, strlen(id)) != 0)
    curmap = curmap->next;

  /* map not found in the stack */
  if (curmap == NULL){
    fprintf(stderr, "map %s search not found in stack\n", id);
    return;
  }

  fprintf(stderr, "map deleting: %s\n", curmap->name);
  tmp = curmap->next;
  free(curmap->name);
  /* base addresses first */

  add = (curmap->format == BRIX || curmap->format == DSN6) ? 1 : 0;
  if (curmap->data){
    for (i=0; i<curmap->len[0]+add; i++){
      for (j=0; j<curmap->len[1]+add; j++)
	free(curmap->data[i][j]);
      free(curmap->data[i]);
    }
    free(curmap->data);
  }

  if (curmap->surface)
    gts_object_destroy(GTS_OBJECT(curmap->surface));
  /* drop map from the linked list */
  free(curmap);

  curmap = tmp;
  /* update map1 */
  map1 = curmap;
}

/* get a z section of a scalar field */
void get_map_plane(gdouble **f, GtsCartesianGrid g, guint k, gpointer *data){
  int i, j;
  double x, y;
  map_ptr *cmap = data[0];
  int *xorg = data[1];
  int *yorg = data[2];
  int *zorg = data[3];

  /* fill up 2D plane with data and return it */
  for (i=0, x=g.x; i<g.nx; i++, x+=g.dx){
    for (j=0, y=g.y; j<g.ny; j++, y+=g.dy){
      f[i][j] = (*cmap)->data[*xorg+i][*yorg+j][*zorg+k];
    }
  }
}

double get_vert_scalar(map_ptr *cmap, GtsVertex2 *v){
  int i, lenx, leny, lenz, gx, gy, gz, nx, ny, nz;
  double x, diffx, y, diffy, z, diffz, gdist;
  double sum1, sum2, scalar;
  double p[8], dist[8];
  vector3 mappt, gridpt, newpt;
  vector3 w, u[8];  

  v3_initialize(&mappt, GTS_POINT(v)->x,
		GTS_POINT(v)->y,
		GTS_POINT(v)->z);
  matrix3_transform(&mappt, (*cmap)->f_t_c);

  /* setup some variables */
  x = (*cmap)->scalars->ori[0];
  diffx = (*cmap)->scalars->diff[0];
  lenx = (*cmap)->scalars->len[0];

  y = (*cmap)->scalars->ori[1];
  diffy = (*cmap)->scalars->diff[1];
  leny = (*cmap)->scalars->len[1];

  z = (*cmap)->scalars->ori[2];
  diffz = (*cmap)->scalars->diff[2];
  lenz = (*cmap)->scalars->len[2];

  /* find closest map point to surface vertex */
  gx = 0; gy = 0; gz = 0;

  gdist = 9999.0;
  for (nx=0; nx < lenx; nx++){
    newpt.x = x + nx * diffx;
    newpt.y = GTS_POINT(v)->y;
    newpt.z = GTS_POINT(v)->z;
    matrix3_transform(&newpt, (*cmap)->scalars->f_t_c);
    if (v3_distance(&newpt, &mappt) < gdist){
      gx = nx;
      gdist = v3_distance(&newpt, &mappt);
    }
  }
  /* out of bounds */
  if (gdist == 9999.0)
    return 0.0;

  gdist = 9999.0;
  for (ny=0; ny < leny; ny++){
    newpt.x = x + gx * diffx;
    newpt.y = y + ny * diffy;
    newpt.z = GTS_POINT(v)->z;
    matrix3_transform(&newpt, (*cmap)->scalars->f_t_c);
    if (v3_distance(&newpt, &mappt) < gdist){
      gy = ny;
      gdist = v3_distance(&newpt, &mappt);
    }
  }
  /* out of bounds */
  if (gdist == 9999.0)
    return 0.0;

  gdist = 9999.0;
  for (nz=0; nz < lenz; nz++){
    newpt.x = x + gx * diffx;
    newpt.y = y + gy * diffy;
    newpt.z = z + nz * diffz;
    matrix3_transform(&newpt, (*cmap)->scalars->f_t_c);
    if (v3_distance(&newpt, &mappt) < gdist){
      gz = nz;
      gdist = v3_distance(&newpt, &mappt);
    }
  }
  /* out of bounds */
  if (gdist == 9999.0)
    return 0.0;

  /* transform the "close" point */
  gridpt.x = x + gx * diffx;
  gridpt.y = y + gy * diffy;
  gridpt.z = z + gz * diffz;
  matrix3_transform(&gridpt, (*cmap)->scalars->f_t_c);

  /* now generate a weighted scalar for the point */
  x = gridpt.x;
  y = gridpt.y;
  z = gridpt.z;
  v3_initialize(&u[0], gridpt.x, gridpt.y, gridpt.z);

  v3_initialize(&u[1], gridpt.x + diffx, gridpt.y, gridpt.z);
  x = diffx; lenx = 1;
  v3_initialize(&newpt, gridpt.x - diffx, gridpt.y, gridpt.z);
  if (v3_distance(&newpt, &mappt) < v3_distance(&u[1], &mappt)){
    u[1].x = newpt.x;
    x = -diffx; lenx = -1;
  }

  v3_initialize(&u[2], gridpt.x, gridpt.y + diffy, gridpt.z);
  y = diffy; leny = 1;
  v3_initialize(&newpt, gridpt.x, gridpt.y - diffy, gridpt.z);
  if (v3_distance(&newpt, &mappt) < v3_distance(&u[2], &mappt)){
    u[2].y = newpt.y;
    y = -diffy; leny = -1;
  }

  v3_initialize(&u[3], gridpt.x, gridpt.y, gridpt.z + diffz);
  z = diffz; lenz = 1;
  v3_initialize(&newpt, gridpt.x, gridpt.y, gridpt.z - diffz);
  if (v3_distance(&newpt, &mappt) < v3_distance(&u[3], &mappt)){
    u[3].z = newpt.z;
    z = -diffz; lenz = -1;
  }

  v3_initialize(&u[4], gridpt.x+x, gridpt.y+y, gridpt.z);
  v3_initialize(&u[5], gridpt.x+x, gridpt.y,   gridpt.z+z);
  v3_initialize(&u[6], gridpt.x,   gridpt.y+y, gridpt.z+z);
  v3_initialize(&u[7], gridpt.x+x, gridpt.y+y, gridpt.z+z);

  if ((gx+lenx*1 < 0) || (gx+lenx*1 >= (*cmap)->scalars->len[0]) ||
      (gy+leny*1 < 0) || (gy+leny*1 >= (*cmap)->scalars->len[1]) ||
      (gz+lenz*1 < 0) || (gz+lenz*1 >= (*cmap)->scalars->len[2])){
    fprintf(stderr, "out of bounds: %d %d %d %d %d %d\n", gx, lenx, gy, leny, gz, lenz);
    return 0.0;
  }

  p[0] = (*cmap)->scalars->data[gx       ][gy       ][gz       ];
  p[1] = (*cmap)->scalars->data[gx+lenx*1][gy       ][gz       ];
  p[2] = (*cmap)->scalars->data[gx       ][gy+leny*1][gz       ];
  p[3] = (*cmap)->scalars->data[gx       ][gy+leny*1][gz+lenz*1];
  p[4] = (*cmap)->scalars->data[gx+lenx*1][gy+leny*1][gz       ];
  p[5] = (*cmap)->scalars->data[gx+lenx*1][gy       ][gz+lenz*1];
  p[6] = (*cmap)->scalars->data[gx       ][gy+leny*1][gz+lenz*1];
  p[7] = (*cmap)->scalars->data[gx+lenx*1][gy+leny*1][gz+lenz*1];

  for (i=0; i<8; i++)
    dist[i] = v3_distance(&mappt, &u[i]);

  for (i=0, sum1=0.0, sum2=0.0; i<8; i++){
    sum1 += dist[i] * p[i];
    sum2 += dist[i];
  }

  scalar = sum1 / sum2;

  return scalar;
}

void sum_area (GtsTriangle *t, gdouble *area){
  int i;
  GtsEdge *e1, *e2, *e3;
  GtsTriangle *tnew;
  GtsVertex *v[3], *w[3];
  vector3 u[3];

  gts_triangle_vertices(t, &v[0], &v[1], &v[2]);

  if (!GTS_VERTEX2(v[0])->draw && !GTS_VERTEX2(v[1])->draw && !GTS_VERTEX2(v[2])->draw)
    return;

  /* transform each vertex */
  for (i=0; i<3; i++){
    u[i].x = GTS_POINT(v[i])->x;
    u[i].y = GTS_POINT(v[i])->y;
    u[i].z = GTS_POINT(v[i])->z;

    matrix3_transform(&u[i], curmap->f_t_c);

    w[i] = gts_vertex_new(gts_vertex_class(), u[i].x, u[i].y, u[i].z);
  }

  e1 = gts_edge_new(gts_edge_class(), GTS_VERTEX(w[0]), GTS_VERTEX(w[1]));
  e2 = gts_edge_new(gts_edge_class(), GTS_VERTEX(w[0]), GTS_VERTEX(w[2]));
  e3 = gts_edge_new(gts_edge_class(), GTS_VERTEX(w[1]), GTS_VERTEX(w[2]));
  tnew = gts_triangle_new(gts_triangle_class(), e1, e2, e3);

  *area += gts_triangle_area(tnew);
}
 
gdouble gts2_surface_area (GtsSurface * s){
  gdouble area = 0.0;

  gts_surface_foreach_face (s, (GtsFunc)sum_area, &area);
  return area;
}

void volume_foreach_face (GtsTriangle *t, gdouble *volume){
  int i;
  GtsVertex *v[3];
  vector3 u[3];

  gts_triangle_vertices(t, &v[0], &v[1], &v[2]);

  if (!GTS_VERTEX2(v[0])->draw && !GTS_VERTEX2(v[1])->draw && !GTS_VERTEX2(v[2])->draw)
    return;

  /* transform each vertex */
  for (i=0; i<3; i++){
    u[i].x = GTS_POINT(v[i])->x;
    u[i].y = GTS_POINT(v[i])->y;
    u[i].z = GTS_POINT(v[i])->z;

    matrix3_transform(&u[i], curmap->f_t_c);
  }

  *volume += (u[0].x * (u[1].y * u[2].z - u[1].z * u[2].y) +
	      u[1].x * (u[2].y * u[0].z - u[2].z * u[0].y) +
	      u[2].x * (u[0].y * u[1].z - u[0].z * u[1].y));
}

gdouble gts2_surface_volume (GtsSurface * s){
  gdouble volume = 0.0;

  g_return_val_if_fail (s != NULL, 0.0);

  gts_surface_foreach_face (s, (GtsFunc) volume_foreach_face, &volume);

  return volume/6.;
}

/* simple file open function */
FILE *gfopen(char *filename, char *mode){
  FILE *fp;

  if ((fp = fopen(filename, mode)) == NULL){
    perror("fopen");
    return NULL;
  }
  return fp;
}

/*
 * determine which vertex is highest and lowest after transform
 * and apply to map
 */
void map_get_low_high(vector3 *low, vector3 *high, vector3 *lowest, vector3 *highest){
  vector3 temp;
  int i, j, k;
 
  lowest->x = 1.0e20;
  lowest->y = 1.0e20;
  lowest->z = 1.0e20;
  highest->x = -1.0e20;
  highest->y = -1.0e20;
  highest->z = -1.0e20;
 
  for(i=0; i<2; i++){
    for(j=0; j<2; j++){
      for(k=0; k<2; k++){
	temp.x = i ? low->x : high->x;
	temp.y = j ? low->y : high->y;
	temp.z = k ? low->z : high->z;

	matrix3_transform(&temp, curmap->c_t_f);

	if (temp.x < lowest->x)
	  lowest->x = temp.x;
	if (temp.y < lowest->y)
	  lowest->y = temp.y;
	if (temp.z < lowest->z)
	  lowest->z = temp.z;
 
	if (temp.x > highest->x)
	  highest->x = temp.x;
	if (temp.y > highest->y)
	  highest->y = temp.y;
	if (temp.z > highest->z)
	  highest->z = temp.z;
      }
    }
  }
}

void map_set_plus(void){
  assert (dstack_size == 1);

  plus_level = dstack[0];
  clear_dstack();
}

void map_xform (double xform[4][4]){
  double temp[4][4], xinv[4][4];
 
  /* transform map given a molscript transform */
  matrix3_invert(xinv, xform);
  for(curmap = map1; curmap; curmap=curmap->next) {
    switch (curmap->format){
    case BRIX:
    case DSN6:
    case APBS:
    case GRASP:
    case GTS:
    case MSMS:
    case MOLSURF:
      matrix3_mult(temp, xform, curmap->f_t_c);
      matrix3_copy(curmap->f_t_c, temp);

      matrix3_mult(temp, curmap->c_t_f, xinv);
      matrix3_copy(curmap->c_t_f, temp);
      break;
    }
  }
}

/* draw the triangles using line segments */
void output_gtsedge(GtsSegment *s, gpointer *data){
  vector3 v1, v2;
  line_segment *ls;
  map_ptr **cmap = data[0];

  if (!GTS_VERTEX2(s->v1)->draw || !GTS_VERTEX2(s->v2)->draw)
    return;

  v1.x = GTS_POINT(s->v1)->x;
  v1.y = GTS_POINT(s->v1)->y;
  v1.z = GTS_POINT(s->v1)->z;
  v2.x = GTS_POINT(s->v2)->x;
  v2.y = GTS_POINT(s->v2)->y;
  v2.z = GTS_POINT(s->v2)->z;

  matrix3_transform(&v1, (**cmap)->f_t_c);
  matrix3_transform(&v2, (**cmap)->f_t_c);

  ls = line_segment_next();
  colour_set_rgb(&ls->c, GTS_VERTEX2(s->v1)->col[0], GTS_VERTEX2(s->v1)->col[1], GTS_VERTEX2(s->v1)->col[2]);
  ls->p = v1;
  ls->new = TRUE;
  ext3d_update(&(ls->p), 0.0);

  ls = line_segment_next();
  colour_set_rgb(&ls->c, GTS_VERTEX2(s->v2)->col[0], GTS_VERTEX2(s->v2)->col[1], GTS_VERTEX2(s->v2)->col[2]);
  ls->p = v2;
  ext3d_update(&(ls->p), 0.0);
}

/* output surface */
void output_gtssurface(map_ptr *cmap){
  boolean colourparts=FALSE;
  gpointer data[2];
  FILE *ofp;

  data[0] = &cmap;
  /* do we want to only update the 3D extent? */
  data[1] = FALSE;

  if (current_state->mapoutmode){
    /* write data here */
    if ((ofp = gfopen(current_state->mapoutfile, "w")) == NULL){
      fprintf(stderr, "output file %s cannot be opened\n", current_state->mapoutfile);
      return;
    }
    fprintf(stderr, "output file for surface %s: \"%s\"\n", (*cmap)->name, current_state->mapoutfile);
    gts_surface_write((*cmap)->surface, ofp);
    current_state->mapoutmode = FALSE;
    fclose(ofp);
  }

  if (output_mode == POVRAY_MODE && current_state->maptype != 0){
    fprintf(stderr, "generating mesh2 (POVRay specific) output for %s\n", (*cmap)->name);
    data[1] = TRUE;
    gts_surface_foreach_face((*cmap)->surface, (GtsFunc) output_gtstri, data);
    output_mesh2(&cmap);
  }
  else if (current_state->maptype == 0){
    fprintf(stderr, "generating line mesh output for %s\n", (*cmap)->name);
    /* so lines merge properly in POVRay */
    if (current_state->transparency != 0.0 && output_mode == POVRAY_MODE)
      output_start_object();

    /* cheap hack to get around colourparts restrictions for lines */
    if ((*cmap)->drawflag == DISTANCES && !current_state->colourparts){
      colourparts = TRUE;
      current_state->colourparts = TRUE;
    }
    line_segment_init();
    gts_surface_foreach_edge((*cmap)->surface, (GtsFunc) output_gtsedge, data);
    output_line(FALSE);
    if (colourparts)
      current_state->colourparts = FALSE;

    if (current_state->transparency != 0.0 && output_mode == POVRAY_MODE)
      output_finish_object();
  }
  else{
    fprintf(stderr, "generating surface mesh output for %s\n", (*cmap)->name);
      if (output_mode == POVRAY_MODE && (current_state->mergemap == 2 || current_state->mergemap == 3))
	;
      else
        output_start_object();

      gts_surface_foreach_face((*cmap)->surface, (GtsFunc) output_gtstri, data);

      /* called twice: once to finish mesh, second to finish merge */
      if (output_mode == POVRAY_MODE && (current_state->mergemap == 0 || current_state->mergemap == 3))
	output_finish_object();

      output_finish_object();
  }
}

/* draw the triangles using object input for molscript */
void output_gtstri(GtsTriangle *t, gpointer *data){
  int i;
  vector3 object[9], tmp[6], u;
  map_ptr **cmap = data[0];
  GtsVertex *v[3];

  boolean set_extent_only = data[1];

  gts_triangle_vertices(t, &v[0], &v[1], &v[2]);

  if (!GTS_VERTEX2(v[0])->draw && !GTS_VERTEX2(v[1])->draw && !GTS_VERTEX2(v[2])->draw)
    return;

  /* transform each vertex, setup object */
  for (i=0; i<3; i++){
    u.x = GTS_POINT(v[i])->x;
    u.y = GTS_POINT(v[i])->y;
    u.z = GTS_POINT(v[i])->z;

    matrix3_transform(&u, (**cmap)->f_t_c);
    object[i*3] = u;

    u.x = GTS_VERTEX2(v[i])->norm[0];
    u.y = GTS_VERTEX2(v[i])->norm[1];
    u.z = GTS_VERTEX2(v[i])->norm[2];

    matrix3_rotate(&u, (**cmap)->f_t_c);
    v3_normalize(&u);

    object[(i*3)+1] = u;

    object[(i*3)+2].x = GTS_VERTEX2(v[i])->col[0];
    object[(i*3)+2].y = GTS_VERTEX2(v[i])->col[1];
    object[(i*3)+2].z = GTS_VERTEX2(v[i])->col[2];
  }

  ext3d_update(&object[0], 0.0);
  ext3d_update(&object[3], 0.0);
  ext3d_update(&object[6], 0.0);

  if (set_extent_only)
    return;

  if (current_state->smoothmap)
    output_object(OBJ_TRIANGLES_NORMALS_COLOURS, object, 9);
  else{
    /* in case we don't want normals */
    tmp[0] = object[0];
    tmp[2] = object[3];
    tmp[4] = object[6];
 
    tmp[1] = object[2];
    tmp[3] = object[5];
    tmp[5] = object[8];
 
    output_object(OBJ_TRIANGLES_COLOURS, tmp, 6);
  }
}

/* remove tris from a surface */
boolean remove_tri(GtsTriangle *t, gpointer *data){
  int i;
  GtsVertex *v[3];

  gts_triangle_vertices(t, &v[0], &v[1], &v[2]);

  for (i=0; i<3; i++){
    if (!GTS_VERTEX2(v[i])->draw)
      return(TRUE);
  }

  return(FALSE);
}

void revert_tri(GtsTriangle *t, gpointer *data){
  gts_triangle_revert(t);
}

void set_coldef_file(char *name){
  fprintf(stderr, "GRASP colour definition file to: %s\n", name);
  curmap->coldef = str_clone(name);
}

/* set color of a vertex based on maptype, etc. */
colour set_colorvert(map_ptr *cmap, float cprop){
  int iprop;
  double weight;
  colour col, maplow, mapmid, maphigh;

  /* initialize col to an *invalid* state */
  col.x = -1.0;
  iprop = (int) cprop;

  if ((*cmap)->drawflag == 0)
    colour_copy_to_rgb(&col, &current_state->linecolour);
  else if (((*cmap)->format != GRASP) &&
	   ((*cmap)->drawflag != CURVATURES) &&
	   ((*cmap)->drawflag != SCALARS)){
    /* setup ramp variables */
    colour_copy_to_rgb(&maplow, &current_state->mapramplow);
    colour_copy_to_rgb(&maphigh, &current_state->mapramphigh);

    mapmid = current_state->mapmidcol;
    if (mapmid.x >= 0.0)
      colour_to_rgb (&mapmid);

    /* colour based drawflag */
    if ((*cmap)->drawflag == POTENTIALS || (*cmap)->drawflag == PROPERTY_2){
      if (iprop >= 0){
	if (((*cmap)->drawflag == POTENTIALS && aa_positive_charge(iprop)) ||
	    ((*cmap)->drawflag == PROPERTY_2 && aa_hbond_donor(iprop)))
	  colour_set_rgb(&col, maplow.x, maplow.y, maplow.z);
	else if (((*cmap)->drawflag == POTENTIALS && aa_negative_charge(iprop)) ||
		 ((*cmap)->drawflag == PROPERTY_2 && aa_hbond_acceptor(iprop)))
	  colour_set_rgb(&col, maphigh.x, maphigh.y, maphigh.z);
      }
      if (!colour_valid_state(&col)){
	if (colour_valid_state(&mapmid))
	  colour_set_rgb(&col, mapmid.x, mapmid.y, mapmid.z);
	else{
	  colour_copy_to_rgb(&col, &maplow);
	  colour_blend(&col, 0.5, &maphigh);
	}
      }
    }
    else if ((*cmap)->drawflag == PROPERTY_1){
      weight = (iprop >= 0) ? aa_kyte_doolittle_normalized(iprop) : 0.5;
      colour_weightedrange_colour(&col, &maplow, &mapmid, &maphigh, weight, 0.0, 0.5, 1.0);
    }
    else{
      /* otherwise make something up */
      colour_copy_to_rgb(&col, &current_state->linecolour);
    }
  }
  else{
    if ((*cmap)->drawflag == VERTEXCOLOR){
      /* GRASP surface with discrete colors */
      if (iprop > 0)
	iprop--;

      /* if out of range, make something up */
      if (iprop > 8)
	colour_copy_to_rgb(&col, &current_state->linecolour);
      else
	colour_set_rgb(&col, (*cmap)->comp[iprop].x, (*cmap)->comp[iprop].y, (*cmap)->comp[iprop].z);
    }
    else{
      /* surface rendering using ranged color property */
      if (current_state->surflow == 0.0 && current_state->surfmid == 0.0 && current_state->surfhigh == 0.0){
	current_state->surflow = (*cmap)->collow;
	current_state->surfmid = (*cmap)->colmid;
	current_state->surfhigh = (*cmap)->colhigh;
      }

      /* setup ramp variables */
      colour_copy_to_rgb(&maplow, &current_state->mapramplow);
      colour_copy_to_rgb(&maphigh, &current_state->mapramphigh);

      /* setup ramp based on GRASP values */
      mapmid = current_state->mapmidcol;
      if (mapmid.x >= 0.0)
	colour_to_rgb (&mapmid);

      colour_weightedrange_colour(&col, &maplow, &mapmid, &maphigh,
			       (double) cprop, current_state->surflow, current_state->surfmid, current_state->surfhigh);
    }
  }

  return(col);
}

/* determine if a triangle should be drawn (but don't remove it) */
void set_drawable(GtsTriangle *t, gpointer *data){
  int i, j;
  double x, y, z;
  boolean draw = FALSE;
  vector3 u[3];
  GtsVertex *v[3];
  GSList *list;

  map_ptr *cmap = data[0];
  int *natoms = data[1];
  at3d ***mapatoms = data[2];
  double radius = current_state->mapradius * current_state->mapradius;

  gts_triangle_vertices(t, &v[0], &v[1], &v[2]);

  for (i=0; i<3; i++){
    u[i].x = GTS_POINT(v[i])->x;
    u[i].y = GTS_POINT(v[i])->y;
    u[i].z = GTS_POINT(v[i])->z;
    matrix3_transform(&u[i], (*cmap)->f_t_c);
  }

  /* if there is a selection */
  if (current_state->mapradius_property){
    for (i=0; i<3; i++){
      if (GTS_VERTEX2(v[i])->cprop < current_state->mapradius)
	draw=TRUE;
    }
  }
  else if (*mapatoms){
    for (i=0; i<3; i++){
      /* check pt is within the given distance to selected atoms */
      for (j=0; j<*natoms; j++){
	if (v3_close(&u[i], &(*mapatoms)[j]->xyz, radius))
	  draw = TRUE;
      }
    }

    /* invert sense if need be */
    draw = current_state->covering ? draw : !draw;
  }
  else
    draw = TRUE;

  /* remove tris based on slab */
  if (backslab > 0.0){
    for (i=0; i<3; i++){
      if (u[i].z < -backslab)
	draw=FALSE;
    }
  }
  if (frontslab > 0.0){
    for (i=0; i<3; i++){
      if (u[i].z > frontslab)
	draw=FALSE;
    }
  }

  if (draw && current_state->mapclip &&
      (u[0].z > 0.0) && (u[1].z > 0.0) && (u[2].z > 0.0))
    draw = FALSE;

  /* remove tris based on normals */
  for (i=0; i<3; i++){
    if (!draw)
      break;

    u[i].x = u[i].y = u[i].z = 0.0;
    for (list=gts_vertex_faces(v[i], (*cmap)->surface, NULL); list; list=list->next){
      gts_triangle_normal(GTS_TRIANGLE((GtsFace *)list->data), &x, &y, &z);
      u[i].x += x;
      u[i].y += y;
      u[i].z += z;
    }
    if (v3_length(&u[i]) == 0.0){
      draw = FALSE;
      break;
    }

    v3_normalize(&u[i]);
    matrix3_rotate(&u[i], (*cmap)->f_t_c);
  }

  for (i=0; i<3; i++)
    GTS_VERTEX2(v[i])->draw = draw;
}

void set_laplacian_lambda (void){
  assert (dstack_size == 1);

  if ((dstack[0] < 0.0) || (dstack[0] > 1.0)) {
    yyerror ("invalid lambda value (must be between 0-1)");
  } else {
    curmap->lambda = dstack[0];
  }
  clear_dstack();
}

void set_mapcontour (void){
  assert (dstack_size == 1);

  contour_level = dstack[0];
  clear_dstack();
}

void set_scalars (char *mapid){
  map_ptr tmp;

  tmp = map1;
  while (strncmp(tmp->name, mapid, strlen(mapid)) != 0){
    if (tmp->next == NULL){
      fprintf(stderr, "map %s not found in stack\n", mapid);
      return;
    }
    tmp = tmp->next;
  }

  curmap->scalars = tmp;
  fprintf(stderr, "setting scalars %s on map %s\n", curmap->scalars->name, curmap->name);
}

/* set extra vertex info */
void set_vertex_attributes(GtsVertex2 *v, gpointer *data){
  int i;
  double val, x, y, z;
  boolean vert_close;
  colour mapcolour;
  vector3 u;
  GSList *list;

  map_ptr *cmap = data[0];
  int *natoms = data[1];
  at3d ***mapatoms = data[2];
  double radius;

  /* initialize mapcolour to *invalid* state */
  mapcolour.x = -1.0;

  if ((*cmap)->format == BRIX ||
      (*cmap)->format == DSN6 ||
      (*cmap)->format == APBS ||
      (*cmap)->format == MSMS)
    radius = current_state->mapradius * current_state->mapradius;
  else
    radius = current_state->probe_rad + (*cmap)->diff[0] + 0.2;

  /* setup normals and cprop */
  if ((*cmap)->format == BRIX ||
      (*cmap)->format == DSN6 ||
      (*cmap)->format == APBS ||
      (*cmap)->format == MSMS ||
      (*cmap)->format == MOLSURF){
    if (*mapatoms){
      u.x = GTS_POINT(GTS_VERTEX(v))->x;
      u.y = GTS_POINT(GTS_VERTEX(v))->y;
      u.z = GTS_POINT(GTS_VERTEX(v))->z;
      matrix3_transform(&u, (*cmap)->f_t_c);

      /* check pt is within the given distance to selected atoms */
      for (i=0; i<*natoms; i++){
	vert_close = FALSE;

	/* gensurf maps use different distance */
	if (((*cmap)->format == BRIX ||
	     (*cmap)->format == DSN6 ||
	     (*cmap)->format == APBS ||
	     (*cmap)->format == MSMS) &&
	    v3_close(&u, &(*mapatoms)[i]->xyz, radius))
	  vert_close = TRUE;
	else if (((*cmap)->format == MOLSURF) &&
		 v3_distance(&(*mapatoms)[i]->xyz, &u)/(radius+(*mapatoms)[i]->radius) <= 1.0)
	  vert_close = TRUE;

	if (vert_close){
	  if ((*cmap)->format != MSMS)
	    v->cprop = (float) aa_ordinal((*mapatoms)[i]->res->type);

	  if ((*cmap)->drawflag == DISTANCES){
	    if (current_state->colourparts)
	      colour_copy_to_rgb(&mapcolour, &(*mapatoms)[i]->res->colour);
	    else
	      colour_copy_to_rgb(&mapcolour, &(*mapatoms)[i]->colour);
	    v->col[0] = mapcolour.x;
	    v->col[1] = mapcolour.y;
	    v->col[2] = mapcolour.z;
	  }
	}
      }
    }

    /* cases that don't depend on distance to atoms */
    if ((*cmap)->drawflag == CURVATURES){
      if (gts_vertex_gaussian_curvature(GTS_VERTEX(v), (*cmap)->surface, &val)){

	if (val > (*cmap)->colhigh)
	  (*cmap)->colhigh = val;
	if (val < (*cmap)->collow)
	  (*cmap)->collow = val;

	v->cprop = (float) val;
      }
    }
    else if ((*cmap)->drawflag == SCALARS)
      v->cprop = (float) get_vert_scalar(cmap, v);

    /* set up normals */
    if ((*cmap)->format != MSMS){
      v->norm[0] = v->norm[1] = v->norm[2] = 0.0;
      for (list=gts_vertex_faces(GTS_VERTEX(v), (*cmap)->surface, NULL); list; list=list->next){
	gts_triangle_normal(GTS_TRIANGLE((GtsFace *)list->data), &x, &y, &z);

	v->norm[0] += x;
	v->norm[1] += y;
	v->norm[2] += z;
      }
    }
  }

  /* setup colors (unless reading in a savefile) */
  if (!colour_valid_state(&mapcolour) && (*cmap)->format != GTS){
    mapcolour = set_colorvert(cmap, v->cprop);

    v->col[0] = mapcolour.x;
    v->col[1] = mapcolour.y;
    v->col[2] = mapcolour.z;
  }
}

/* setup colour data to read in/set */
void set_verttype(char *type){
  int i;
  char *name[] = {"potential","curvature","distance","property1","property2","discretecolours","globalcolor","scalars",NULL};

  if (strncmp(type, "none", strlen("none")) == 0)
    curmap->drawflag = 0;
  else{
    curmap->drawflag = 1;
    /*
     * loop looks funny in order to jump to the
     * right position in the file later on (result of GRASP format, mostly)
     */
    for (i=0; name[i]; i++){
      if (!(strncmp(type, name[i], strlen(name[i])) == 0))
	curmap->drawflag *= 2;
      else
	break;
    }
    if (!name[i]){
      fprintf(stderr, "surface type %s not found\n", type);
      return;
    }
  }
  #ifdef EDEBUG
  fprintf(stderr, "type: %s drawflag: %d\n", type, curmap->drawflag);
  #endif
}

void smooth_vertex(GtsVertex *v, gpointer *data){
  GtsSurface *s = data[0];
  double lambda = curmap->lambda;

  if (!gts_vertex_is_boundary (v, s)) {
    GSList *vertices = gts_vertex_neighbors (v, NULL, s);
    GSList *i;
    GtsVector U0 = { 0., 0., 0.};
    guint n = 0;
                                                                                
    i = vertices;
    while (i) {
      GtsPoint * p = i->data;
      U0[0] += p->x;
      U0[1] += p->y;
      U0[2] += p->z;
      n++;
      i = i->next;
    }
    g_slist_free (vertices);
                                                                                
    if (n > 0) {
      GTS_POINT (v)->x += (lambda)*(U0[0]/n - GTS_POINT (v)->x);
      GTS_POINT (v)->y += (lambda)*(U0[1]/n - GTS_POINT (v)->y);
      GTS_POINT (v)->z += (lambda)*(U0[2]/n - GTS_POINT (v)->z);
    }
  }
}

/* stop function for simplification (% of vertices removed) */
boolean stop_func(gdouble cost, guint number, gpointer *data){
  double *val = (double *)data;

  /* fprintf(stderr, "%d %f %d\n", number, *val, (guint) (*val)); */
  if (*val > 0 && number < (guint) *val)
    return TRUE;
  else if (*val < 0 && number > (guint) (-1.0 * (*val)))
    return TRUE;
  else
    return FALSE;
}

/* recursive byte swap routine (don't leave home without it) */
void swap(char *x, int size){
  unsigned char c;
  unsigned short s;
  unsigned long l;

  switch (size){
    case 2: /* swap two bytes */
      c = *x;
      *x = *(x+1);
      *(x+1) = c;
      break;
    case 4: /* swap two shorts (2-byte words) */
      s = *(unsigned short *)x;
      *(unsigned short *)x = *((unsigned short *)x + 1);
      *((unsigned short *)x + 1) = s;
      swap ((char *)x, 2);
      swap ((char *)((unsigned short *)x+1), 2);
      break;
    case 8: /* swap two longs (4-bytes words) */
      l = *(unsigned long *)x;
      *(unsigned long *)x = *((unsigned long *)x + 1);
      *((unsigned long *)x + 1) = l;
      swap ((char *)x, 4);
      swap ((char *)((unsigned long *)x+1), 4);
      break;
  }
}

/*
 * dump a chunk of the map to df3 format for input to something like PovRay
 */
void write_df3(at3d ***mapatoms, int atom_count, int xorg, int xext, int yorg, int yext, int zorg, int zext){
  int i, j, k, x;
  int draw_pt;
  int mlt = 1;
  unsigned short buf[3], scaleval;
  float val, min=0.0, max=0.0;
  float radius;
  char outdf3[100];
  FILE *ofp;
  vector3 object, extent, origin;

  /* in case the contour is negative */
  if (contour_level < 0.0)
    mlt = -1;

  if (current_state->dfnum > 999)
    return;
  sprintf(outdf3, "edmap.%d.df3", current_state->dfnum++);
  buf[0] = xext-xorg;
  buf[1] = yext-yorg;
  buf[2] = zext-zorg;

  /* check we can open the file */
  if ((ofp = gfopen(outdf3, "wb")) == NULL){
    fprintf(stderr, "df3 output file %s cannot be opened\n", outdf3);
    return;
  }

  /* offset + origin */
  origin.x = (double) (xorg * curmap->diff[0]) + curmap->ori[0];
  origin.y = (double) (yorg * curmap->diff[1]) + curmap->ori[1];
  origin.z = (double) (zorg * curmap->diff[2]) + curmap->ori[2];
  /* len + origin */
  extent.x = (double) (xext * curmap->diff[0]) + curmap->ori[0];
  extent.y = (double) (yext * curmap->diff[1]) + curmap->ori[1];
  extent.z = (double) (zext * curmap->diff[2]) + curmap->ori[2];
  extent.x -= origin.x;
  extent.y -= origin.y;
  extent.z -= origin.z;

  /* header must be high order byte first */
  #ifndef WORDS_BIGENDIAN
  swap((char *) &buf[0], sizeof(short));
  swap((char *) &buf[1], sizeof(short));
  swap((char *) &buf[2], sizeof(short));
  #endif

  fwrite(buf, sizeof(short), 3, ofp);

  /* df3 is x fast, y medium, z slow */
  for (i=zorg; i<zext; i++){
    for (j=yorg; j<yext; j++){
      for (k=xorg; k<xext; k++){
	/* transform coord */
	object.x = (double) (k * curmap->diff[0]) + curmap->ori[0];
	object.y = (double) (j * curmap->diff[1]) + curmap->ori[1];
	object.z = (double) (i * curmap->diff[2]) + curmap->ori[2];

	/* put coord in realspace terms */
	matrix3_transform(&object, curmap->f_t_c);
	/* should we be drawing it? */
	if (*mapatoms){
	  radius = current_state->mapradius * current_state->mapradius;
	  for (x=0, draw_pt=0; x<atom_count; x++){
	    if (v3_close(&object, &(*mapatoms)[x]->xyz, radius) == TRUE)
	      draw_pt=1;
	  }
	}
	else
	  draw_pt=1;

	/* if so, consider it in min/max calc */
	if (draw_pt){
	  val = mlt * curmap->data[k][j][i];
	  if (val < min)
	    min = val;
	  if (val > max)
	    max = val;
	}
      }
    }
  }

  /* now generate the df3 file */
  for (i=zorg; i<zext; i++){
    for (j=yorg; j<yext; j++){
      for (k=xorg; k<xext; k++){
	object.x = (double) (k * curmap->diff[0]) + curmap->ori[0];
	object.y = (double) (j * curmap->diff[1]) + curmap->ori[1];
	object.z = (double) (i * curmap->diff[2]) + curmap->ori[2];

	matrix3_transform(&object, curmap->f_t_c);
	if (*mapatoms){
	  radius = current_state->mapradius * current_state->mapradius;
	  for (x=0, draw_pt=0; x<atom_count; x++){
	    if (v3_close(&object, &(*mapatoms)[x]->xyz, radius) == TRUE)
	      draw_pt=1;
	  }
	}
	else
	  draw_pt=1;

	/* scale contour relative to max contour in the region */
	val = mlt * curmap->data[k][j][i];
	if (draw_pt && (val >= mlt * contour_level))
	  scaleval = (unsigned short) 255 * (val/max);
	else
	  scaleval = 0;

	fwrite((void *) &scaleval, sizeof(char), 1, ofp);
      }
    }
  }

  fclose(ofp);

  /* now port data */
  output_df3(outdf3, contour_level, mlt * max, &origin, &extent, curmap->f_t_c);

  fprintf(stderr, "df3 contour min: %f contour level: %f max: %f\n", mlt * min, contour_level, mlt * max);
}
