/* povray.c

   MolScript v2.1

   Povray; input for the 'render' program.

   Developed by: Ezra Peisach
    17-Sep-1997  fairly finished
    02-Apr-1998  cleanup for temporary release
    31-Aug-1999  Declare internal variables static and 
                 output some #Declare statements with a semicolon.
    28-Jan-2002  TimF - aniso B support / output tris as a mesh
    09-Apr-2002  TimF - tweak GR_MESH output (maybe add a new output style with a single texture?)
    11-Jul-2002  TimF - POVRay 3.5 mods (camera tweaks to keep RHS)
    09-Mar-2003  TimF - mesh2 output support (GRASP/MSMS only)
    12-Apr-2004  TimF - expand mesh2 support for use with GTS
*/

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include <gts.h>

#include "clib/matrix3.h"
#include "clib/str_utils.h"

#include "povray.h"
#include "global.h"
#include "graphics.h"
#include "segment.h"
#include "state.h"
#include "map_utils.h"
#include "vertex_class.h"

#ifndef M_PI
#define M_PI 3.14159265
#endif

/*------------------------------------------------------------*/
static FILE *proper_outfile;

static int triangle_count;
static int sphere_count;
static int cylinder_count;

typedef struct {
  colour rgb;
  double transp;
  boolean isrgb;
  boolean interior;
} colour_db_entry;


/* Current working state */
static colour_db_entry pstate = {{ COLOUR_RGB, 1.0, 1.0, 1.0}, 0.0};

#define INDENT_MAX 10
static int indent_level = 0;
static char indent_str[INDENT_MAX+1] = "";

static colour_db_entry *colour_array = NULL;
static int colour_alloc = 0;
static int colour_count = 0;
static int mesh_count = 0;

/* for grouping/merging various objects */
enum{ GR_GROUP_OR_MERGE, GR_TANSPARENT, GR_MESH, GR_BLOB };

static int mesh_output = 0;
static int light_count;
#define MAX_LIGHTS 7
static int light_type[MAX_LIGHTS];
static float light_position[MAX_LIGHTS][3];
static colour light_colour[MAX_LIGHTS];
static float light_spot_direction[MAX_LIGHTS][3];

#define MAX_COLNAME 10
static char *colprefix = NULL;

/*------------------------------------------------------------*/
static void
convert_colour (colour *c)
{
  assert (c);

  pstate.rgb = *c;
  colour_to_rgb (&pstate.rgb);
}


/*------------------------------------------------------------*/
static int
add_cache_colour (boolean setrgb, boolean interior)
{
  int i;

  /* Returns a colour that has been cached */

  /* Check cache */
  for(i = 0; i < colour_count; i++) {
    if (!colour_unequal(&pstate.rgb, &colour_array[i].rgb)
	&& (pstate.transp == colour_array[i].transp)
	&& (setrgb == colour_array[i].isrgb)
	&& (interior == colour_array[i].interior)) return i;
  }
  
  /* Add entry to cache */
  if(colour_count >= colour_alloc) {
    colour_alloc *= 2;
    colour_array = realloc (colour_array, colour_alloc * sizeof (colour_db_entry));
    assert(colour_array);
  }
  colour_array[colour_count].rgb = pstate.rgb;
  colour_array[colour_count].transp = pstate.transp;
  colour_array[colour_count].isrgb = setrgb;
  colour_array[colour_count].interior = interior;

  return colour_count++;
 
}

 
/*------------------------------------------------------------*/
static void
write_colour (void)
{
  int cache_line;

  /* Add colour to cache */
  cache_line = mesh_output ? add_cache_colour(FALSE, FALSE) : add_cache_colour(FALSE, TRUE);

  if (mesh_output)
    fprintf(outfile, "texture { %s_%d }", colprefix, cache_line+1);
  else
    fprintf(outfile, "material { %s_%d }", colprefix, cache_line+1);

}

/*------------------------------------------------------------*/
static void
write_rgb_colour (void)
{
  int cache_line;

  /* Add colour to cache */
  cache_line = add_cache_colour(TRUE, FALSE);

  fprintf(outfile, "%s_%d", colprefix, cache_line+1);

}

/*------------------------------------------------------------*/
/* Sets the indention string to indent_level */
static void
set_indent (void)
{
  int i;
  assert(indent_level <= INDENT_MAX);

  for(i=0; i < indent_level; i++) {
    indent_str[i]=' ';
  }

  indent_str[i] = '\0';
}

/*------------------------------------------------------------*/
/* The grouping using merge is to ensure that transparent objects
 * don't have internal planes
 */
static void
start_group (int when)
{
  if(current_state->transparency != 0.0) {
    if ((when == GR_GROUP_OR_MERGE) || (when == GR_TANSPARENT)) {
      fprintf (outfile, "%smerge {\n", indent_str);
      indent_level += 2;
    }
  } else {
    if (when == GR_GROUP_OR_MERGE) {
      fprintf (outfile, "%sunion {\n", indent_str);
      indent_level += 2;
    }
  }

  if (when == GR_MESH){
    fprintf (outfile, "%sobject {\n", indent_str);
    fprintf (outfile, "%s  mesh {\n", indent_str);
    mesh_count++;
    indent_level += 4;
  }
  else if (when == GR_BLOB){
    fprintf (outfile, "%sblob {\n", indent_str);
    indent_level += 2;
  }
                                                                                
  set_indent();
}

/*------------------------------------------------------------*/
static void
end_group (int when)
{
  if ((when == GR_GROUP_OR_MERGE) ||
      ( (when == GR_TANSPARENT) && (current_state->transparency != 0.0))) {
    indent_level -= 2;
    set_indent();
    fprintf (outfile, "%s}\n", indent_str);
  }
  else if (when == GR_MESH){
    indent_level -= 4;
    set_indent();
    fprintf (outfile, "%s  }\n", indent_str);
    fprintf (outfile, "%s  interior { %s_MESH%d }\n", indent_str, colprefix, mesh_count);
    fprintf (outfile, "%s}\n", indent_str);
  }
  else if (when == GR_BLOB){
    fprintf (outfile, "%sthreshold 0.05\n", indent_str);
    indent_level -= 2;
    set_indent();
    fprintf (outfile, "%s}\n", indent_str);
  }
}


/*------------------------------------------------------------*/
static int
write_triangle (vector3 *p1, vector3 *p2, vector3 *p3)
{
  assert (p1);
  assert (p2);
  assert (p3);

  if ((backslab > 0.0) &&
      ((p1->z < -backslab) && (p2->z < -backslab) && (p3->z < -backslab)))
    return FALSE;
  if ((frontslab > 0.0) &&
      ((p1->z > frontslab) && (p2->z > frontslab) && (p3->z > frontslab)))
    return FALSE;

  if (triangle_count == 0)
    pov_start_object();

  fprintf (outfile,
	   "%striangle {<%.3f, %.3f, %.3f>, <%.3f, %.3f, %.3f>, <%.3f, %.3f, %.3f>\n",
	   indent_str,
	   p1->x, p1->y, p1->z,
	   p2->x, p2->y, p2->z,
	   p3->x, p3->y, p3->z);

  fprintf (outfile, "%s  ", indent_str);
  write_colour();

  fprintf (outfile, "\n%s}\n", indent_str);

  triangle_count++;

  return TRUE;
}


/*------------------------------------------------------------*/
static int
write_norm_triangle (vector3 *p1, vector3 *p2, vector3 *p3, 
		     vector3 *n1, vector3 *n2, vector3 *n3)
{
  assert (p1);
  assert (p2);
  assert (p3);
  assert (n1);
  assert (n2);
  assert (n3);

  if ((backslab > 0.0) &&
      ((p1->z < -backslab) && (p2->z < -backslab) && (p3->z < -backslab)))
    return FALSE;
  if ((frontslab > 0.0) &&
      ((p1->z > frontslab) && (p2->z > frontslab) && (p3->z > frontslab)))
    return FALSE;

  if (triangle_count == 0)
    pov_start_object();

  fprintf (outfile, "%ssmooth_triangle {\n", indent_str);

  fprintf (outfile, "%s  <%.3f, %.3f, %.3f>, <%.3f, %.3f, %.3f>,\n", 
	   indent_str, p1->x, p1->y, p1->z, n1->x, n1->y, n1->z);
  fprintf (outfile, "%s  <%.3f, %.3f, %.3f>, <%.3f, %.3f, %.3f>,\n", 
	   indent_str, p2->x, p2->y, p2->z, n2->x, n2->y, n2->z);
  fprintf (outfile, "%s  <%.3f, %.3f, %.3f>, <%.3f, %.3f, %.3f>\n", 
	   indent_str, p3->x, p3->y, p3->z, n3->x, n3->y, n3->z);
  fprintf (outfile, "%s  ", indent_str);
  write_colour();

  fprintf (outfile, "\n%s}\n", indent_str);

  triangle_count++;

  return TRUE;
}


/*------------------------------------------------------------*/
static void

write_anisou_sphere_int (vector3 *p, double evect[3][3], double anisou[3][3], double ainv[3][3], double molmat[4][4], int hasanisou, int output_col)
{
  float crit;
  double A, B, C, D, E, F, J;
  
  assert (p);

  crit = pcrit[(int) (100*current_state->prob_level-1)][1];

  if (current_state->bsurf == 2){
    A = ainv[0][0];
    B = ainv[1][1];
    C = ainv[2][2];
    D = ainv[0][1] * 2.0;
    E = ainv[0][2] * 2.0;
    F = ainv[1][2] * 2.0;

    J = -1.0 * pow(crit, 2.0);

    #ifdef EDEBUG
    fprintf(stderr, "quadric:\n");
    fprintf(stderr, "%f %f %f %f %f %f  %f\n\n", A, B, C, D, E, F, J);
    #endif

    /* ellipsoid quadric */
    fprintf (outfile, "%squadric { \n", indent_str);
    fprintf (outfile, "%s  <%.6f, %.6f, %.6f>,\n%s  <%.6f, %.6f, %.6f>,\n%s  <0.00, 0.00, 0.00>,\n%s  %f\n%s  ", indent_str, A, B, C, indent_str, D, E, F, indent_str, indent_str, J, indent_str);

    /* apply (final) molscript coordinate system */
    fprintf (outfile, "matrix <%.6f, %.6f, %.6f,\n%s  ", molmat[0][0], molmat[0][1], molmat[0][2], indent_str);
    fprintf (outfile, "        %.6f, %.6f, %.6f,\n%s  ", molmat[1][0], molmat[1][1], molmat[1][2], indent_str);
    fprintf (outfile, "        %.6f, %.6f, %.6f,\n%s  ", molmat[2][0], molmat[2][1], molmat[2][2], indent_str);
    fprintf (outfile, "        0.00, 0.00, 0.00>\n%s  ", indent_str);

    /* point to draw ellipsoid */
    fprintf (outfile, "translate <%.2f, %.2f, %.2f>\n%s  ", p->x, p->y, p->z, indent_str);

    /*
     * add some transparency to ellipse so its written
     * as a different colour than the axes
     */
    pstate.transp += current_state->btransp;
    if (pstate.transp > 1.0)
      pstate.transp -= 0.2;
    if (output_col) {
      write_colour();
    }
    fprintf (outfile, "\n%s}\n", indent_str);
  }

  sphere_count++;
}

/*------------------------------------------------------------*/
static void

write_anisou_axes_int (vector3 *p, double evect[3][3], double anisou[3][3], double molmat[4][4], int hasanisou, int output_col)
{
  int i;
  float crit;

  crit = pcrit[(int) (100*current_state->prob_level-1)][1];

  if (current_state->baxes == 1){
    fprintf (outfile, "%smerge {\n%s  ", indent_str, indent_str);
    /* for all 3 eigenvectors */
    for (i=0; i<3; i++){
      fprintf (outfile, "cylinder { ");
      /* cylinder in (+) eigenvector direction */
      fprintf (outfile, "<0.00, 0.00, 0.00>, <%.6f, %.6f, %.6f>, %.3f }\n%s  ", crit*sqrt(anisou[i][i])*evect[0][i], crit*sqrt(anisou[i][i])*evect[1][i], crit*sqrt(anisou[i][i])*evect[2][i], current_state->axesrad, indent_str);
      /* cylinder in (-) eigenvector direction */
      fprintf (outfile, "cylinder { ");
      fprintf (outfile, "<0.00, 0.00, 0.00>, <%.6f, %.6f, %.6f>, %.3f }\n%s  ", -1.0*crit*sqrt(anisou[i][i])*evect[0][i], -1.0*crit*sqrt(anisou[i][i])*evect[1][i], -1.0*crit*sqrt(anisou[i][i])*evect[2][i], current_state->axesrad, indent_str);
    }

    /* molscript matrix */
    fprintf (outfile, "matrix <%.6f, %.6f, %.6f,\n%s  ", molmat[0][0], molmat[0][1], molmat[0][2], indent_str);
    fprintf (outfile, "        %.6f, %.6f, %.6f,\n%s  ", molmat[1][0], molmat[1][1], molmat[1][2], indent_str);
    fprintf (outfile, "        %.6f, %.6f, %.6f,\n%s  ", molmat[2][0], molmat[2][1], molmat[2][2], indent_str);
    fprintf (outfile, "        0.00, 0.00, 0.00>\n%s  ", indent_str);

    /* point to draw cylinder */
    fprintf (outfile, "translate <%.2f, %.2f, %.2f>\n%s  ", p->x, p->y, p->z, indent_str);

    if (output_col) {
      write_colour();
    }
    fprintf (outfile, "\n%s}\n", indent_str);
  }
  else{
    fprintf (outfile, "%sunion {\n", indent_str);
    fprintf (outfile, "%s  disc { \n", indent_str);
    /* disc 1 */
    fprintf (outfile, "%s    <0.00, 0.00, 0.00>,\n%s    <1.00, 0.00, 0.00>,\n%s    1.0 }\n", indent_str, indent_str, indent_str);
    /* disc 2 (normal in y) */
    fprintf (outfile, "%s  disc { \n", indent_str);
    fprintf (outfile, "%s    <0.00, 0.00, 0.00>,\n%s    <0.00, 1.00, 0.00>,\n%s    1.0 }\n", indent_str, indent_str, indent_str);
    /* disc 3 (normal in z) */
    fprintf (outfile, "%s  disc { \n", indent_str);
    fprintf (outfile, "%s    <0.00, 0.00, 0.00>,\n%s    <0.00, 0.00, 1.00>,\n%s    1.0 }\n", indent_str, indent_str, indent_str);

    /* scale by eigenvalues */
    fprintf (outfile, "%s  scale <%.6f, %.6f, %.6f>\n%s  ", indent_str, crit*sqrt(anisou[0][0]), crit*sqrt(anisou[1][1]), crit*sqrt(anisou[2][2]), indent_str);

    /* set up coordinate system based on eigenvectors */
    fprintf (outfile, "matrix <%.6f, %.6f, %.6f,\n%s  ", evect[0][0], evect[1][0], evect[2][0], indent_str);
    fprintf (outfile, "        %.6f, %.6f, %.6f,\n%s  ", evect[0][1], evect[1][1], evect[2][1], indent_str);
    fprintf (outfile, "        %.6f, %.6f, %.6f,\n%s  ", evect[0][2], evect[1][2], evect[2][2], indent_str);
    fprintf (outfile, "        0.00, 0.00, 0.00>\n%s  ", indent_str);

    /* now set up coordinate system based on molscript */
    fprintf (outfile, "matrix <%.6f, %.6f, %.6f,\n%s  ", molmat[0][0], molmat[0][1], molmat[0][2], indent_str);
    fprintf (outfile, "        %.6f, %.6f, %.6f,\n%s  ", molmat[1][0], molmat[1][1], molmat[1][2], indent_str);
    fprintf (outfile, "        %.6f, %.6f, %.6f,\n%s  ", molmat[2][0], molmat[2][1], molmat[2][2], indent_str);
    fprintf (outfile, "        0.00, 0.00, 0.00>\n%s  ", indent_str);

    /* and translate to point to actually draw stuff */
    fprintf (outfile, "translate <%.2f, %.2f, %.2f>\n%s  ", p->x, p->y, p->z, indent_str);
    if (output_col) {
      write_colour();
    }
    fprintf (outfile, "\n%s}\n", indent_str);
  }
}

/*------------------------------------------------------------*/
static void

write_sphere_int (vector3 *p, double radius, int output_col)
{
  assert (p);
  assert (radius > 0.0);

  fprintf (outfile,
	   "%ssphere { <%.2f, %.2f, %.2f>, %.2f  ",
	   indent_str, p->x, p->y, p->z, radius);
  if (output_col) {
    write_colour();
  }
  fprintf (outfile, " }\n");

  sphere_count++;
}

/*------------------------------------------------------------*/
static void
write_anisou_sphere (vector3 *p, double evect[3][3], double anisou[3][3], double ainv[3][3], double molmat[4][4], int hasanisou)
{
  write_anisou_sphere_int (p, evect, anisou, ainv, molmat, hasanisou, TRUE);
}

/*------------------------------------------------------------*/
static void
write_anisou_axes (vector3 *p, double evect[3][3], double anisou[3][3], double molmat[4][4], int hasanisou)
{
  write_anisou_axes_int (p, evect, anisou, molmat, hasanisou, TRUE);
}

/*------------------------------------------------------------*/
static void
write_sphere (vector3 *p, double radius)
{
  write_sphere_int (p, radius, TRUE);
}

/*------------------------------------------------------------*/
static void
write_sphere_nocol (vector3 *p, double radius)
{
  write_sphere_int (p, radius, FALSE);
}

/*------------------------------------------------------------*/
static void
write_cylinder_round (vector3 *p1, vector3 *p2, double radius, boolean group)
{
  vector3 vec;

  assert (p1);
  assert (p2);
  assert (radius > 0.0);

  if ((backslab > 0.0) &&
      ((p1->z - radius < -backslab) && (p2->z - radius < -backslab)))
    return;
  if ((frontslab > 0.0) &&
      ((p1->z + radius > frontslab) && (p2->z + radius > frontslab)))
    return;

  /* Degenerate cylinders are bad in povray
     so remove distances less then 0.01A away. 
     We do not use the distance formula as it can mask a rounding error 
   */
  v3_difference(&vec, p1, p2);
  if((fabs(vec.x) < 0.01) && (fabs(vec.y) < 0.01) && (fabs(vec.z) < 0.01)) {
    /* Should we show anything? */
    return;
  }

  if (group && cylinder_count == 0)
    start_group(GR_TANSPARENT);
  
  start_group(GR_GROUP_OR_MERGE);
  fprintf (outfile,
	   "%scylinder {<%.2f,%.2f,%.2f>, <%.2f, %.2f, %.2f>, %.2f}\n",
	   indent_str, p1->x, p1->y, p1->z, p2->x, p2->y, p2->z, radius);
  write_sphere_nocol(p1, radius);
  write_sphere_nocol(p2, radius);
  fprintf (outfile, "%s", indent_str);
  write_colour();
  fprintf (outfile, "\n");
  end_group(GR_GROUP_OR_MERGE);

  cylinder_count++;
}


/*------------------------------------------------------------*/
static void
write_cylinder_flat (vector3 *p1, vector3 *p2, double radius, boolean group)
{
  assert (p1);
  assert (p2);
  assert (radius > 0.0);

  if ((backslab > 0.0) &&
      ((p1->z - radius < -backslab) && (p2->z - radius < -backslab)))
    return;
  if ((frontslab > 0.0) &&
      ((p1->z + radius > frontslab) && (p2->z + radius > frontslab)))
    return;

  if (group && cylinder_count == 0)
    start_group(GR_TANSPARENT);

  fprintf (outfile,
	   "%scylinder {<%.2f,%.2f,%.2f>, <%.2f, %.2f, %.2f>, %.2f\n",
	   indent_str, p1->x, p1->y, p1->z, p2->x, p2->y, p2->z, radius);
  fprintf (outfile, "%s  ", indent_str);
  write_colour();
  fprintf (outfile, "%s}\n", indent_str);
  cylinder_count++;
}

/*------------------------------------------------------------*/
static void
write_material (void)
{

  pstate.transp = current_state->transparency;
#if 0
  if ((current_state->shininess == material_shininess) &&
      (current_state->transparency == material_transparency) &&
      ! colour_unequal (&(current_state->specularcolour),
			&(material_specularcolour))) return;

  if (special_material) fprintf (outfile, "9\n");

  material_shininess = current_state->shininess;
  material_transparency = current_state->transparency;
  colour_copy (&(material_specularcolour), &(current_state->specularcolour));

  convert_colour (&(material_specularcolour));
  fprintf (outfile, "8\n%.2g %.3g %.3g %.3g %.3g %.3g 1 0 0 0\n",
	   material_shininess * 128.0,
	   (material_transparency == 0.0) ? 0.25 : 0.6,
	   rgb.x, rgb.y, rgb.z,
	   material_transparency);

  special_material = TRUE;
#endif
}

/*------------------------------------------------------------*/
static void
write_line (vector3 *p1, vector3 *p2)
{
  double radius = LINEWIDTH_FACTOR * current_state->linewidth;

  assert (p1);
  assert (p2);

  if (radius < LINEWIDTH_MINIMUM) radius = LINEWIDTH_MINIMUM;

  if (current_state->linedash == 0.0) {
    write_cylinder_round (p1, p2, radius, FALSE);

  } else {
    int slot, parts;
    vector3 pos1, pos2;

    parts = (int) (v3_distance (p1, p2) * 16.0 / current_state->linedash);
    if (parts <= 1) {
      write_cylinder_round (p1, p2, radius, FALSE);

    } else if ((current_state->linewidth != 0.0) &&
	       (current_state->linedash / current_state->linewidth < 2.0)) {
      parts /= 2;
      for (slot= 0; slot <= parts; slot++) {
	v3_between (&pos1, p1, p2, (double) slot / (double) parts);
	write_sphere (&pos1, radius);
      }

    } else {
      if (parts % 2 == 0) parts++; /* last segment should end on last point */
      write_sphere (p1, radius);
      for (slot = 0; slot < parts; slot += 2) {
	v3_between (&pos1, p1, p2, (double) slot / (double) parts);
	v3_between (&pos2, p1, p2, (double) (slot + 1) / (double) parts);
	write_cylinder_flat (&pos1, &pos2, radius, FALSE);
      }
      write_sphere (p2, radius);
    }
  }
}


/*------------------------------------------------------------*/
void
pov_set (void)
{
  output_first_plot = pov_first_plot;
  output_start_plot = pov_start_plot;
  output_finish_plot = pov_finish_plot;
  output_finish_output = pov_finish_output;

  set_area = pov_set_area;
  set_background = pov_set_background;
  anchor_start = do_nothing_str;
  anchor_description = do_nothing_str;
  anchor_parameter = do_nothing_str;
  anchor_start_geometry = do_nothing;
  anchor_finish = do_nothing;
  lod_start = do_nothing;
  lod_finish = do_nothing;
  lod_start_group = do_nothing;
  lod_finish_group = do_nothing;
  viewpoint_start = do_nothing_str;
  viewpoint_output = do_nothing;

  output_directionallight = pov_directionallight;
  output_pointlight = pov_pointlight;
  output_spotlight = do_nothing;
  output_comment = pov_comment;

  output_coil = pov_coil;
  output_cylinder = pov_cylinder;
  output_df3 = pov_df3;
  output_helix = pov_helix;
  output_label = pov_label;
  output_line = pov_line;
  output_mesh2 = pov_mesh2;
  output_sphere = pov_sphere;
  output_stick = pov_stick;
  output_strand = pov_strand;

  output_start_object = pov_start_object;
  output_object = pov_object;
  output_finish_object = pov_fin_object;

  output_start_blob = pov_start_blob;
  output_blob = pov_blob;
  output_finish_blob = pov_fin_blob;

  output_pickable = NULL;

  constant_colours_to_rgb();

  output_mode = POVRAY_MODE;
}


/*------------------------------------------------------------*/
void
pov_first_plot (void)
{
  if (!first_plot)
    yyerror ("only one plot per input file allowed for Povray");

  set_outfile ("w");

  proper_outfile = outfile;
  if (tmp_filename == NULL) {
    outfile = tmpfile();
  } else {
    outfile = fopen (tmp_filename, "w+");
  }
  if (outfile == NULL) yyerror ("could not create the temporary output file");

  return;
}


/*------------------------------------------------------------*/
void
pov_finish_output (void)
{
  int c, i, slot;
  colour *col;
  FILE *tmp_outfile;

  tmp_outfile = outfile;
  outfile = proper_outfile;

  /* 
   * Prepend anything to the top of the output file, i.e. colour cache 
   */

  /* Insert background */

  /* General settings */

  PRINT("#ifndef (NoHeaders)\n");
  if(input_filename) fprintf(outfile, "// Input file: %s\n", input_filename);
  fprintf(outfile, "// Size input as %dx%d\n", output_width, output_height);
  fprintf(outfile, "camera { orthographic\n");
  fprintf(outfile, "         location <0, 0, %.2f>\n", window*2.0);
  PRINT("         direction <0, 0, -1>\n");
  fprintf(outfile,"         up %.2f*y\n", window*2.0);
  if (output_width == output_height)
    fprintf(outfile,"         right %.2f*x\n", window*2.0);
  else if (2*output_width == 3*output_height) 
    fprintf(outfile,"         right 3/2*%.2f*x\n", window*2.0);
  else if (3*output_width == 2*output_height) 
    fprintf(outfile,"         right 2/3*%.2f*x\n", window*2.0);
  else if (3*output_width == 4*output_height) 
    fprintf(outfile,"         right 4/3*%.2f*x\n", window*2.0);
  else if (4*output_width == 3*output_height) 
    fprintf(outfile,"         right 3/4*%.2f*x\n", window*2.0);
  else
    fprintf(outfile, "         right %d/%d*%.2f*x\n", output_width, output_height, window*2.0);
  PRINT("         look_at <0, 0, 0> }\n\n");

  if (headlight) {
    PRINT("light_source { <10, 10, 50> color rgb .6 }\n");
    PRINT("light_source { <10, 10, 50> color rgb .6 shadowless }\n\n");
  }

  /* Render the other lights */
  if (light_count) {
    for (slot = 0 ; slot < light_count; slot++) {
      
    }
  }

  /* For backwards compatibility, if lightambientintensity is not set, do not 
     output any ambient keywords */
  if(current_state->lightambientintensity != 0.0) {
    fprintf(outfile, "global_settings { ambient_light rgb %.3f }\n",
	    current_state->lightambientintensity);
  }


  fprintf(outfile, "background { color rgb < %.3f, %.3f, %.3f>}\n",
	  background_colour.x, background_colour.y, background_colour.z);

  if (fog != 0.0) {
    /* constant fog */
    fprintf(outfile, "fog {fog_type 1 distance %.3f color rgb < %.3f, %.3f, %.3f>}\n",
	    fog,
	  background_colour.x, background_colour.y, background_colour.z);
  }

  PRINT("#default { texture { finish { ");
  if(current_state->lightambientintensity != 0.0) {
    PRINT("ambient 1 ");
  }
  PRINT("specular 0.5 roughness 0.02 } } }\n");

  PRINT("\n#declare NoHeaders = 1;\n");
  PRINT("#end\n\n");

  PRINT("#ifndef (MyClear)\n");
  PRINT("#declare MyClear = rgbf 1;\n");
  PRINT("#end\n");
  PRINT("#ifndef (RFont)\n");
  PRINT("#declare RFont=\"timrom.ttf\"\n");
  PRINT("#end\n");
  PRINT("#ifndef (GFont)\n");
  PRINT("#declare GFont=\"symbol.ttf\"\n");
  PRINT("#end\n");
  PRINT("#ifndef (FontWidth)\n");
  PRINT("#declare FontWidth=0.2;\n");
  PRINT("#end\n");
  PRINT("\n");

  /* Insert colour cache */
  for (i = 0 ; i < colour_count; i++) {
    col = &colour_array[i].rgb;
    fprintf(outfile, "#ifndef (%s_%d)\n", colprefix, i+1);

    if (colour_array[i].isrgb){
      fprintf(outfile, "#declare %s_%d = ",
	      colprefix, i+1);
    }
    else{
      if (colour_array[i].interior){
	fprintf(outfile, "#declare %s_%d = material { texture { pigment ",
		colprefix, i+1);
      }
      else{
	fprintf(outfile, "#declare %s_%d = texture { pigment ",
		colprefix, i+1);
      }
    }

    if (colour_array[i].isrgb){
      if(colour_array[i].transp != 0.0) {
	fprintf(outfile, "rgbt <%.3f, %.3f, %.3f, %.3f>;\n", 
		col->x, col->y, col->z, colour_array[i].transp);
      } else {
	fprintf(outfile, "rgb <%.3f, %.3f, %.3f>;\n", 
		col->x, col->y, col->z);
      }
    }
    else{
      if(colour_array[i].transp != 0.0) {
	fprintf(outfile, "{rgbt <%.3f, %.3f, %.3f, %.3f>}",
		col->x, col->y, col->z, colour_array[i].transp);
      } else {
	fprintf(outfile, "{rgb <%.3f, %.3f, %.3f>}",
		col->x, col->y, col->z);
      }
      fprintf(outfile, "%s}\n", colour_array[i].interior ? " }" : " ");

    }

    fprintf(outfile, "#end\n");
  }

  if (mesh_count > 0){
    for (i=0; i<mesh_count; i++){
      fprintf(outfile, "#ifndef (%s_MESH%d)\n", colprefix, i+1);
      fprintf(outfile, "#declare %s_MESH%d = interior { }\n", colprefix, i+1);
      fprintf(outfile, "#end\n");
    }
  }


  /*
   * Copy temp file to output
   */
  fseek (tmp_outfile, 0L, SEEK_SET);
  while ((c = fgetc (tmp_outfile)) != EOF) fputc (c, outfile);
  fclose (tmp_outfile);
  if (tmp_filename != NULL) remove (tmp_filename);
  return;
}


/*------------------------------------------------------------*/
void
pov_start_plot (void)
{
  triangle_count = 0;
  sphere_count = 0;
  cylinder_count = 0;
  light_count = 0;

  /* Sets a reasonable aspect ratio - simply for guessing */
  set_area_values(-output_width/2, -output_height/2, output_width/2, output_height/2);
  background_colour = black_colour;

  /* Initialize the colour cache */
  if(colour_alloc == 0) {
    colour_alloc = 16;
    colour_array = malloc (colour_alloc * sizeof (colour_db_entry));
  }
  colour_count = 0;

  if(input_filename) {
    char *ptr;
    colprefix = str_clone(input_filename);
    for(ptr = colprefix; *ptr; ptr++) {
      if(islower(*ptr)) *ptr=toupper(*ptr);
      if(*ptr == '.') {
	*ptr = 0;
	break;
      }
    }
    if(strlen(colprefix) > MAX_COLNAME) {
      colprefix[MAX_COLNAME-2] = colprefix[strlen(colprefix)-2];
      colprefix[MAX_COLNAME-1] = colprefix[strlen(colprefix)-1];
      colprefix[MAX_COLNAME]='\0';
    }
  }
  else 
    colprefix = str_clone("POV_COL");

}


/*------------------------------------------------------------*/
void
pov_finish_plot (void)
{


  /* Get the extents */
  set_extent();

  if (message_mode)
    fprintf (stderr, "%i triangles, %i spheres and %i cylinders.\n",
	     triangle_count, sphere_count, cylinder_count);
}


/*------------------------------------------------------------*/
void
pov_set_area (void)
{
  if (message_mode) fprintf (stderr, "ignoring 'area' for Povray output\n");
  clear_dstack();
}


/*------------------------------------------------------------*/
void
pov_set_background (void)
{
  colour_to_rgb(&given_colour);
  background_colour = given_colour;
}


/*------------------------------------------------------------*/
void
pov_directionallight (void)
{
  vector3 lightdirection;

  assert ((dstack_size == 3) || (dstack_size == 6));

  if (dstack_size == 3) {
    lightdirection.x = dstack[0];
    lightdirection.y = dstack[1];
    lightdirection.z = dstack[2];
  } else {
    lightdirection.x = dstack[3] - dstack[0];
    lightdirection.y = dstack[4] - dstack[1];
    lightdirection.z = dstack[5] - dstack[2];
  }
  v3_normalize (&lightdirection);

  convert_colour (&(current_state->lightcolour));
  fprintf (outfile, "light_source {<%.2f, %.2f, %.2f>*100000 ",
 	   lightdirection.x, lightdirection.y, lightdirection.z);
  fprintf (outfile, "color rgb <%.2f, %.2f, %.2f>", 
	   pstate.rgb.x, pstate.rgb.y, pstate.rgb.z);
  if (!shadows) 
    PRINT(" shadowless");
  PRINT("}\n");


  clear_dstack();
}


/*------------------------------------------------------------*/
void
pov_pointlight (void)
{
  assert (dstack_size == 3);

  convert_colour (&(current_state->lightcolour));
  fprintf (outfile, "light_source {<%.2f, %.2f, %.2f> ",
 	   dstack[0], dstack[1], dstack[2]);
  fprintf (outfile, "color rgb <%.2f, %.2f, %.2f>", 
	   pstate.rgb.x, pstate.rgb.y, pstate.rgb.z);
  if (!shadows) 
    PRINT(" shadowless");
  PRINT("}\n");
  clear_dstack();
}


/*------------------------------------------------------------*/
void
pov_comment (char *str)
{
  assert (str);

  fprintf (outfile, "// %s\n", str);
}


/*------------------------------------------------------------*/
void
pov_coil (void)
{
  int slot;
  coil_segment *cs1, *cs2;

  write_material();

  cylinder_count = 0;

  if (current_state->colourparts) {

    for (slot = 1; slot < coil_segment_count; slot++) {
      cs1 = coil_segments + slot - 1;
      cs2 = coil_segments + slot;
      convert_colour (&(cs1->c));
      write_cylinder_round (&(cs1->p), &(cs2->p), current_state->coilradius, TRUE);
    }

  } else {
    convert_colour (&(current_state->planecolour));

    for (slot = 1; slot < coil_segment_count; slot++) {
      cs1 = coil_segments + slot - 1;
      cs2 = coil_segments + slot;
      write_cylinder_round (&(cs1->p), &(cs2->p), current_state->coilradius, TRUE);
    }
  }

  if (cylinder_count > 0){
    cylinder_count += coil_segment_count;
    fprintf(stderr, "%i cylinders in coil\n", cylinder_count);
    end_group(GR_TANSPARENT);
  }
  cylinder_count = 0;

}


/*------------------------------------------------------------*/
void
pov_cylinder (vector3 *v1, vector3 *v2)
{
  assert (v1);
  assert (v2);
  assert (v3_distance (v1, v2) > 0.0);

  write_material();
  convert_colour (&(current_state->planecolour));
  write_cylinder_flat (v1, v2, current_state->cylinderradius, FALSE);
}

/*------------------------------------------------------------*/
void
pov_df3 (char *outdf3, float min, float max, vector3 *v1, vector3 *v2, double f_t_c[4][4])
{
  colour low, mid, high;

  assert (v1);
  assert (v2);

  write_material();

  /* just dump all the df3 info to the povray output file */
  /* scaling info */
  if (current_state->mergemap == 0)
    fprintf(outfile, "%s // density is scaled from %f to %f\n", indent_str, min, max);
  /*
   * are we merging df3 files?
   * if not, start a new "box"
   */
  if ((current_state->mergemap == 0) || (current_state->mergemap == 1)){
    fprintf(outfile, "%sbox{\n", indent_str);
    fprintf(outfile, "%s <0,0,0>, <1,1,1>\n", indent_str);
    fprintf(outfile, "%s texture{ pigment {rgbt 1} finish {ambient rgb <0,0,0>}}\n", indent_str);
    fprintf(outfile, "%s interior{\n", indent_str);
  }
  /* POVRay settings */
  if (current_state->mergemap != 0)
    fprintf(outfile, "%s   // density is scaled from %f to %f\n", indent_str, min, max);
  fprintf(outfile, "%s   media{\n", indent_str);
  fprintf(outfile, "%s      intervals 100\n", indent_str);
  fprintf(outfile, "%s      samples 1,20\n", indent_str);
  fprintf(outfile, "%s      emission <0.5,0.5,0.5>\n", indent_str);
  fprintf(outfile, "%s      absorption <0,0,0>\n", indent_str);
  fprintf(outfile, "%s      scattering { 1,<0,0,0> }\n", indent_str);
  fprintf(outfile, "%s      confidence 0.999\n", indent_str);
  fprintf(outfile, "%s      variance 1/500\n", indent_str);
  fprintf(outfile, "%s      density {\n", indent_str);
  fprintf(outfile, "%s         density_file df3 \"%s\" interpolate 1\n", indent_str, outdf3);
  fprintf(outfile, "%s         density_map {\n", indent_str);
  fprintf(outfile, "%s            [0 rgb 0]\n", indent_str);
  fprintf(outfile, "%s            [1.0 rgb 10]\n", indent_str);
  fprintf(outfile, "%s         }\n", indent_str);
  fprintf(outfile, "%s         color_map {\n", indent_str);
  fprintf(outfile, "%s            [0.0 rgb <0,0,0>]\n", indent_str);

  /* colour setup */

  colour_copy_to_rgb(&low, &current_state->mapramplow);
  convert_colour (&low);
  fprintf(outfile, "%s            [0.001 ", indent_str);
  write_rgb_colour();
  fprintf(outfile, "]\n");

  if (current_state->mapmidcol.x >= 0.0){
    colour_copy_to_rgb(&mid, &current_state->mapmidcol);
    convert_colour(&mid);
    fprintf(outfile, "%s            [0.5 ", indent_str);
    write_rgb_colour();
    fprintf(outfile, "]\n");
  }

  colour_copy_to_rgb(&high, &current_state->mapramphigh);
  convert_colour(&high);
  fprintf(outfile, "%s            [1.0 ", indent_str);
  write_rgb_colour();
  fprintf(outfile, "]\n");

  fprintf(outfile, "%s         }\n", indent_str);
  fprintf(outfile, "%s      }\n", indent_str);
  fprintf(outfile, "%s   }\n", indent_str);
  /*
   * are we done?
   * if so, end with transform parameters
   * (this won't work if maps are merged across two different areas)
   */
  if ((current_state->mergemap == 0) || (current_state->mergemap == 3)){
    fprintf(outfile, "%s }\n", indent_str);
    fprintf(outfile, "%s no_shadow\n", indent_str);
    fprintf(outfile, "%s hollow\n\n", indent_str);
    fprintf(outfile, "%s scale <%f, %f, %f>\n", indent_str, v2->x, v2->y, v2->z);
    fprintf(outfile, "%s translate <%f, %f, %f>\n", indent_str, v1->x, v1->y, v1->z);
    fprintf(outfile, "%s matrix <%f, %f, %f,\n", indent_str, f_t_c[0][0], f_t_c[0][1], f_t_c[0][2]);
    fprintf(outfile, "%s         %f, %f, %f,\n", indent_str, f_t_c[1][0], f_t_c[1][1], f_t_c[1][2]);
    fprintf(outfile, "%s         %f, %f, %f,\n", indent_str, f_t_c[2][0], f_t_c[2][1], f_t_c[2][2]);
    fprintf(outfile, "%s         %f, %f, %f>\n", indent_str, f_t_c[3][0], f_t_c[3][1], f_t_c[3][2]);
    fprintf(outfile, "%s}\n", indent_str);
  }
}

/*------------------------------------------------------------*/
void
pov_mesh2 (map_ptr **cmap)
{
  gint i, nface, nvert;
  gint repeat_alloc = 16;
  gint count, start_col, ncol;
  gint repeat_count;
  gint cache_line;
  gint *col_array;
  gint *col_repeat = malloc(repeat_alloc * sizeof(gint));
  boolean norm, count_gts;
  gpointer data[8];

  write_material();

  if (current_state->mergemap == 1){
    fprintf(outfile, "%smerge {\n", indent_str);
    indent_level += 2;
    set_indent();
  }

  fprintf(outfile, "%sobject {\n", indent_str);
  indent_level += 2;
  set_indent();
  fprintf(outfile, "%smesh2 {\n", indent_str);
  mesh_count++;

  /* get # verts */
  i = 0;
  norm = FALSE;
  count_gts = TRUE;
  data[0] = &cmap;
  data[1] = &i;
  data[2] = &norm;
  data[3] = &count_gts;
  gts_surface_foreach_vertex((**cmap)->surface, (GtsFunc) pov_mesh2_vertex, data);
  nvert = i;

  /* vertex vectors */
  fprintf(outfile, "%s vertex_vectors {\n", indent_str);
  fprintf(outfile, "%s  %u,\n", indent_str, nvert);

  /* now write verts */
  i = 0;
  count_gts = FALSE;
  gts_surface_foreach_vertex((**cmap)->surface, (GtsFunc) pov_mesh2_vertex, data);

  if (i % 3)
    fprintf(outfile, "\n%s }\n", indent_str);
  else
    fprintf(outfile, "%s }\n", indent_str);

  if (current_state->smoothmap){
    /* normal vectors */
    fprintf(outfile, "%s normal_vectors {\n", indent_str);
    fprintf(outfile, "%s  %u,\n", indent_str, nvert);

    /* now write normals */
    i = 0;
    norm = TRUE;
    gts_surface_foreach_vertex((**cmap)->surface, (GtsFunc) pov_mesh2_vertex, data);

    if (i % 3)
      fprintf(outfile, "\n%s }\n", indent_str);
    else
      fprintf(outfile, "%s }\n", indent_str);
  }

  /* get # colours */
  /* we'll need to save color indices for later */
  if ((col_array = calloc(nvert, sizeof(gint))) == NULL){
    perror("calloc");
    exit(1);
  }

  /* first find NEW colours */
  i = ncol = repeat_count = 0;
  count = start_col = colour_count;
  data[0] = col_array;
  data[1] = &start_col;
  data[2] = &i;
  data[3] = &count;
  data[4] = &ncol;
  data[5] = col_repeat;
  data[6] = &repeat_alloc;
  data[7] = &repeat_count;
  gts_surface_foreach_vertex((**cmap)->surface, (GtsFunc) pov_mesh2_get_newcolindex, data);

  /* now find predefined colours */
  i = 0;
  gts_surface_foreach_vertex((**cmap)->surface, (GtsFunc) pov_mesh2_get_oldcolindex, data);

  if ((**cmap)->drawflag != 0){
    /* texture list */
    fprintf(outfile, "%s texture_list {\n", indent_str);
    fprintf(outfile, "%s  %u,\n", indent_str, ncol + repeat_count);

    /* now write colours */
    for (i=start_col; i<start_col+ncol; i++)
      fprintf(outfile, "%s  texture { %s_%u },\n", indent_str, colprefix, i+1);

    /* add any additional colours (i.e. already defined colours) */
    for (i=0; i<repeat_count; i++)
      fprintf(outfile, "%s  texture { %s_%d },\n", indent_str, colprefix, col_repeat[i]+1);

    fprintf(outfile, "%s }\n", indent_str);

    free(col_repeat);
  }

  /* face indices */
  i = 0;
  norm = FALSE;
  count_gts = TRUE;
  data[0] = col_array;
  data[1] = &i;
  data[2] = &norm;
  data[3] = &count_gts;
  data[4] = GINT_TO_POINTER((**cmap)->drawflag == 0 ? FALSE : TRUE);
  gts_surface_foreach_face((**cmap)->surface, (GtsFunc) pov_mesh2_face, data);
  nface = i;

  fprintf(outfile, "%s face_indices {\n", indent_str);
  fprintf(outfile, "%s  %u,\n", indent_str, nface);

  /* now write faces (+color) indices */
  i = 0;
  count_gts = FALSE;
  gts_surface_foreach_face((**cmap)->surface, (GtsFunc) pov_mesh2_face, data);

  if (i % 3)
    fprintf(outfile, "\n%s }\n", indent_str);
  else
    fprintf(outfile, "%s }\n", indent_str);

  if (current_state->smoothmap){
    /* normal indices */
    fprintf(outfile, "%s normal_indices {\n", indent_str);

    /* get # normals */
    fprintf(outfile, "%s  %u,\n", indent_str, nface);

    /* now write normal indices */
    i = 0;
    norm = TRUE;
    gts_surface_foreach_face((**cmap)->surface, (GtsFunc) pov_mesh2_face, data);

    if (i % 3)
      fprintf(outfile, "\n%s }\n", indent_str);
    else
      fprintf(outfile, "%s }\n", indent_str);
  }

  fprintf(outfile, "%s}\n", indent_str);
  if ((**cmap)->drawflag == 0){
    for (i=start_col; i<start_col+ncol; i++)
      fprintf(outfile, "%stexture { %s_%u }\n", indent_str, colprefix, i+1);

    /* add any additional colours (i.e. already defined colours) */
    for (i=0; i<repeat_count; i++)
      fprintf(outfile, "%stexture { %s_%d }\n", indent_str, colprefix, col_repeat[i]+1);

    if (current_state->interiorcolour.x != -1.0){
      convert_colour(&(current_state->interiorcolour));
      cache_line = add_cache_colour(FALSE, FALSE);
      fprintf(outfile, "%sinterior_texture { %s_%d }\n", indent_str, colprefix, cache_line+1);
    }

    free(col_repeat);
  }
  fprintf(outfile, "%sinterior { %s_MESH%d }\n", indent_str, colprefix, mesh_count);
  indent_level -= 2;
  set_indent();
  fprintf(outfile, "%s}\n", indent_str);

  if (current_state->mergemap == 3){
    indent_level -= 2;
    set_indent();
    fprintf(outfile, "%s}\n", indent_str);
  }

  /* done with mesh2 */
  fprintf(stderr, "%u verts %u colours and %u faces in mesh2 output\n", nvert, ncol, nface);

  gts_surface_foreach_vertex((**cmap)->surface, (GtsFunc) gts_object_reset_reserved, NULL);

  free(col_array);
}

/* output face information for mesh2 */
void pov_mesh2_face(GtsTriangle *t, gpointer *data){
  gint a, b, c;
  GtsVertex *v1, *v2, *v3;
  gint *col_array = data[0];
  gint *i = data[1];
  boolean *norm = data[2];
  boolean *count = data[3];
  boolean drawflag = (boolean) data[4];

  gts_triangle_vertices(t, &v1, &v2, &v3);
  if (!GTS_VERTEX2(v1)->draw && !GTS_VERTEX2(v2)->draw && !GTS_VERTEX2(v3)->draw)
    return;

  if (*count){
    *i+=1;
    return;
  }

  a = GPOINTER_TO_UINT(GTS_OBJECT(v1)->reserved);
  b = GPOINTER_TO_UINT(GTS_OBJECT(v2)->reserved);
  c = GPOINTER_TO_UINT(GTS_OBJECT(v3)->reserved);

  if (!(*i % 3))
    fprintf(outfile, "%s  ", indent_str);

  if (*norm || !drawflag)
    fprintf(outfile, "<%u, %u, %u>, ", a, b, c);
  else
    fprintf(outfile, "<%u, %u, %u>, %d, %d, %d, ", a, b, c, col_array[a], col_array[b], col_array[c]);

  *i+=1;
  if (!(*i % 3))
    fprintf(outfile, "\n");
}

/* determine color indices not already in cache for mesh2 */
void pov_mesh2_get_newcolindex(GtsVertex2 *v, gpointer *data){
  gint col_index;
  colour col;
  gint *col_array = data[0];
  gint *start_col = data[1];
  gint *i = data[2];
  gint *count = data[3];
  gint *ncol = data[4];

  if (!v->draw)
    return;

  colour_set_rgb(&col, v->col[0], v->col[1], v->col[2]);
  convert_colour(&col);

  col_index = add_cache_colour(FALSE, FALSE);

  if (*count != (gint) colour_count){
    *count = (gint) colour_count;
    col_array[*i] = *ncol;
    *ncol+=1;
  }
  else{
    /* its NOT a new colour but WAS defined in this mesh */
    col_array[*i] = col_index - *start_col;
  }

  *i+=1;
}

/* cross reference colors */
void pov_mesh2_get_oldcolindex(GtsVertex2 *v, gpointer *data){
  gint j, col_index;
  boolean hitcol;
  colour col;
  gint *col_array = data[0];
  gint *start_col = data[1];
  gint *i = data[2];
  gint *count = data[3];
  gint *ncol = data[4];
  gint *col_repeat = data[5];
  gint *repeat_alloc = data[6];
  gint *repeat_count = data[7];

  if (!v->draw)
    return;

  colour_set_rgb(&col, v->col[0], v->col[1], v->col[2]);
  convert_colour(&col);

  col_index = add_cache_colour(FALSE, FALSE);

  if (col_index - *start_col < 0){
    /* its NOT a new colour and WASN'T defined in this mesh */
    /* so we need to keep track of which ones */
    hitcol = FALSE;
    if (*repeat_count > 0){
      /* did we already add it? */
      for (j=0; j<*repeat_count; j++){
	if (col_repeat[j] == col_index){
	  hitcol = TRUE;
	  col_array[*i] = *ncol+j;
	  break;
	}
      }
    }

    if (!hitcol){
      /* if not, add it to the list */
      if (*repeat_count > *repeat_alloc){
	*repeat_alloc *= 2;
	col_repeat = realloc(col_repeat, *repeat_alloc * sizeof(gint));
	assert(*col_repeat);
      }
      col_repeat[*repeat_count] = col_index;

      col_array[*i] = (*ncol + *repeat_count);

      *repeat_count+=1;
    }
  }
  *i+=1;
}

/* output vertex information for mesh2 */
void pov_mesh2_vertex(GtsVertex2 *v, gpointer *data){
  map_ptr ***cmap = data[0];
  gint *i = data[1];
  boolean *norm = data[2];
  boolean *count = data[3];
  vector3 u;

  if (!v->draw)
    return;

  if (*count){
    *i+=1;
    return;
  }

  if (*norm){
    u.x = v->norm[0];
    u.y = v->norm[1];
    u.z = v->norm[2];
    matrix3_rotate(&u, (***cmap)->f_t_c);
  }
  else{
    u.x = GTS_POINT(v)->x;
    u.y = GTS_POINT(v)->y;
    u.z = GTS_POINT(v)->z;
    matrix3_transform(&u, (***cmap)->f_t_c);
  }

  if (!(*i % 3))
    fprintf(outfile, "%s  ", indent_str);

  fprintf(outfile, "<%.3f, %.3f, %.3f>, ", u.x, u.y, u.z);

  GTS_OBJECT(v)->reserved = GUINT_TO_POINTER (*i);
  *i+=1;

  if (!(*i % 3))
    fprintf(outfile, "\n");
}

/*------------------------------------------------------------*/
static void 
calc_normal(vector3 *out, vector3 *p1, vector3 *p2, vector3 *p21, vector3 *p22)
{
  vector3 mid, v1, v2;

  v3_middle(&mid, p21, p22);
  v3_difference(&v1, p1, &mid);
  v3_difference(&v2, p2, &mid);
  v3_cross_product(out, &v1, &v2);
  v3_normalize(out);
}
  
/*------------------------------------------------------------*/
void
pov_start_blob (void)
{
  start_group(GR_BLOB);
}

/*------------------------------------------------------------*/
void
pov_blob (at3d *at, double radius)
{
  assert (at);
  assert (radius > 0.0);

  /* cheat in order to use texture, not material */
  mesh_output=1;

  write_material();
  convert_colour (&(at->colour));

  fprintf (outfile,
	   "%ssphere { <%.2f, %.2f, %.2f>, %.2f, 1.0  ",
	   indent_str, at->xyz.x, at->xyz.y, at->xyz.z, radius);
  write_colour();
  fprintf (outfile, " }\n");

  mesh_output=0;
  sphere_count++;
}

/*------------------------------------------------------------*/
void
pov_fin_blob (void)
{
  end_group(GR_BLOB);
}

/*------------------------------------------------------------*/
void
pov_helix (void)
{
  int slot;
  helix_segment *hs1, *hs2;
  vector3 p1, p2, p21, p22;
  vector3 ip1, ip2, ip21, ip22;
  vector3 norm, prevnorm, norm2, prevnorm2;

  double offset;

  mesh_output=1;

  write_material();
  convert_colour (&(current_state->planecolour));

  for (slot = 1; slot < helix_segment_count; slot++) { /* outer flat side */
    hs1 = helix_segments + slot - 1;
    hs2 = hs1 + 1;
    if (current_state->colourparts) convert_colour (&(hs1->c));
    write_norm_triangle (&(hs1->p1), &(hs1->p2), &(hs2->p1),
			 &(hs1->n), &(hs1->n), &(hs2->n));
    write_norm_triangle (&(hs1->p2), &(hs2->p1), &(hs2->p2),
			 &(hs1->n), &(hs2->n), &(hs2->n));
  }


  if (current_state->helixthickness > 0.01) {
    offset = current_state->helixthickness;
  } else {
    offset = 0.02;
  }

  if (current_state->helixthickness > 0.01) {
    hs1 = helix_segments;	/* first segment edge */
    if (current_state->colourparts) convert_colour (&(hs1->c));
    v3_sum_scaled (&ip1, &(hs1->p1), -offset, &(hs1->n));
    v3_sum_scaled (&ip2, &(hs1->p2), -offset, &(hs1->n));
    write_triangle(&(hs1->p1), &(hs1->p2), &ip1);
    write_triangle(&(hs1->p2), &ip2, &ip1);


    for (slot = 1; slot < helix_segment_count; slot++) { /* top and  */
      hs1 = helix_segments + slot - 1;                   /* bottom edges */
      hs2 = hs1 + 1;
      if (current_state->colourparts) convert_colour (&(hs1->c));
      v3_sum_scaled (&ip1, &(hs1->p1), -offset, &(hs1->n));
      v3_sum_scaled (&ip2, &(hs1->p2), -offset, &(hs1->n));
      v3_sum_scaled (&ip21, &(hs2->p1), -offset, &(hs2->n));
      v3_sum_scaled (&ip22, &(hs2->p2), -offset, &(hs2->n));
      /* top edge */
      calc_normal(&norm, &ip1, &(hs1->p1), &ip21, &(hs2->p1));
      if(slot == 1) prevnorm = norm;

      write_norm_triangle(&(hs1->p1), &(hs2->p1), &ip1,
			  &prevnorm, &norm, &prevnorm);
      write_norm_triangle(&(hs2->p1), &ip21, &ip1,
			  &norm, &norm, &prevnorm);
      prevnorm = norm;

      /* bottom edge */
      calc_normal(&norm2, &(hs1->p2), &ip2, &(hs2->p2), &ip22);
      if(slot == 1) prevnorm2 = norm2;
      write_norm_triangle(&(hs1->p2), &(hs2->p2), &ip2,
			  &prevnorm2, &norm2, &prevnorm2);
      write_norm_triangle(&(hs2->p2), &ip22, &ip2,
			  &norm2, &norm2, &prevnorm2);
      prevnorm2 = norm2;
    }

    hs1 = helix_segments + helix_segment_count - 1; /* last segment edge */
    if (current_state->colourparts) convert_colour (&(hs1->c));
    v3_sum_scaled (&ip1, &(hs1->p1), -offset, &(hs1->n));
    v3_sum_scaled (&ip2, &(hs1->p2), -offset, &(hs1->n));
    write_triangle(&(hs1->p1), &(hs1->p2), &ip1);
    write_triangle(&(hs1->p2), &ip2, &ip1);
  }

  convert_colour (&(current_state->plane2colour));

  for (slot = 0; slot < helix_segment_count; slot++) {
    v3_reverse (&(helix_segments + slot)->n);
  }

  for (slot = 1; slot < helix_segment_count; slot++) { /* inner flat side */
    hs1 = helix_segments + slot - 1;
    hs2 = hs1 + 1;
    if (current_state->colourparts) convert_colour (&(hs1->c));
    v3_sum_scaled (&p1, &(hs1->p1), offset, &(hs1->n));
    v3_sum_scaled (&p2, &(hs1->p2), offset, &(hs1->n));
    v3_sum_scaled (&p21, &(hs2->p1), offset, &(hs2->n));
    v3_sum_scaled (&p22, &(hs2->p2), offset, &(hs2->n));
    write_norm_triangle (&p1, &p2, &p21,
			 &(hs1->n), &(hs1->n), &(hs2->n));
    write_norm_triangle (&p2, &p21, &p22,
			 &(hs1->n), &(hs2->n), &(hs2->n));
  }

  if (triangle_count > 0)
    pov_fin_object();
}


/*------------------------------------------------------------*/
void
pov_label (vector3 *p, char *label, colour *c)
{
  vector3 pos;
  double size = current_state->labelsize/15.0; /* How did we get 15? */

  assert (p);
  assert (label);
  assert (*label);

  write_material();
  if (c) {
    convert_colour (c);
  } else {
    convert_colour (&(current_state->linecolour));
  }

  v3_sum (&pos, p, &(current_state->labeloffset));

  fprintf(outfile, "text { ttf RFont, \n");
  fprintf(outfile, "  \"%s\", FontWidth, 0\n", label);
  fprintf(outfile, "  scale <%.2f, %.2f, 1>\n", size, size);
  fprintf(outfile, "  translate -y*%.2f\n", size*0.75/2.0);
  if(current_state->labelrotation == TRUE) 
    fprintf(outfile, "  rotate <0, 0, 90>\n");
  fprintf(outfile, "  translate <%.2f, %.2f, %.2f>\n", pos.x, pos.y, pos.z);
  write_colour();
  fprintf(outfile, " }\n");


}


/*------------------------------------------------------------*/
void
pov_line (boolean polylines)
{
  int slot;
  line_segment *ls1, *ls2;

  if (line_segment_count < 2) return;

  write_material();

  if (current_state->colourparts) {
    
    if (polylines) {

      for (slot = 1; slot < line_segment_count; slot++) {
	ls1 = line_segments + slot - 1;
	ls2 = line_segments + slot;
	if (ls2->new) continue;
	convert_colour (&(ls1->c));
	write_line (&(ls1->p), &(ls2->p));
      }

    } else {
      for (slot = 0; slot < line_segment_count; slot += 2) {
	ls1 = line_segments + slot;
	ls2 = line_segments + slot + 1;
	convert_colour (&(ls1->c));
	write_line (&(ls1->p), &(ls2->p));
      }
    }

  } else {
    convert_colour (&(line_segments[0].c));

    if (polylines) {
      for (slot = 1; slot < line_segment_count; slot++) {
	ls1 = line_segments + slot - 1;
	ls2 = line_segments + slot;
	if (ls2->new) continue;
	write_line (&(ls1->p), &(ls2->p));
      }

    } else {
      for (slot = 0; slot < line_segment_count; slot += 2) {
	ls1 = line_segments + slot;
	ls2 = line_segments + slot + 1;
	write_line (&(ls1->p), &(ls2->p));
      }
    }
  }

}


/*------------------------------------------------------------*/
void
pov_sphere (at3d *at, double radius)
{
  int i, j;
  double evect[3][3], anisou[3][3];
  double inverse[3][3];

  assert (at);
  assert (radius > 0.0);

  if ((backslab > 0.0) && (at->xyz.z - radius < -backslab))
    return;
  if ((frontslab > 0.0) && (at->xyz.z + radius > frontslab))
    return;

  write_material();
  convert_colour (&(at->colour));

  if (current_state->bsurf > 0){

    #ifdef EDEBUG
    if (! at->hasaniso)
      fprintf(stderr, "isotropic values: %f  Ueq: %f\n", at->bfactor, at->bfactor / (8.0 * M_PI * M_PI));

    fprintf(stderr, "U:\n");
    for (i=0; i<3; i++){
      for (j=0; j<3; j++){
	fprintf(stderr, "%f ", at->anisou[i][j]);
      }
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");
    #endif

    for (i=0; i<3; i++)
      for (j=0; j<3; j++)
        anisou[i][j] = at->anisou[i][j];

    matrix2_invert (inverse, anisou);
    matrix2_diagonalize(evect, anisou);

    #ifdef EDEBUG
    fprintf(stderr, "inverse matrix:\n");
    for (i=0; i<3; i++){
      for (j=0; j<3; j++){
	fprintf(stderr, "%f ", inverse[i][j]);
      }
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "eigenvectors:\n");
    fprintf(stderr, "(column major order)\n");
    for (i=0; i<3; i++){
      for (j=0; j<3; j++){
	fprintf(stderr, "%f ", evect[i][j]);
      }
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "diagonalized U:\n");
    for (i=0; i<3; i++){
      for (j=0; j<3; j++){
	fprintf(stderr, "%f ", anisou[i][j]);
      }
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");
    #endif

    for (i=0; i<3; i++){
      if (anisou[i][i] <= 0.00){
        fprintf(stderr, " *** residue %s atom %s: non-positive definite (won't be drawn) *** \n", at->res->name, at->name);
        return;
      }
    }

    if (at->hasaniso){
      if (current_state->baxes > 0)
	write_anisou_axes (&(at->xyz), evect, anisou, at->res->mol->molmat, TRUE);
      if (current_state->bsurf > 1)
	write_anisou_sphere (&(at->xyz), evect, anisou, inverse, at->res->mol->molmat, TRUE);
    }
    else{
      if (current_state->baxes > 0)
	write_anisou_axes (&(at->xyz), evect, anisou, at->res->mol->molmat, FALSE);
      if (current_state->bsurf > 1)
	write_anisou_sphere (&(at->xyz), evect, anisou, inverse, at->res->mol->molmat, FALSE);
    }
  }
  else
    write_sphere (&(at->xyz), radius);
}


/*------------------------------------------------------------*/
void
pov_stick (vector3 *v1, vector3 *v2, double r1, double r2, colour *c)
{
  assert (v1);
  assert (v2);
  assert (v3_distance (v1, v2) > 0.0);

  write_material();
  if (c) {
    convert_colour (c);
  } else {
    convert_colour (&(current_state->planecolour));
  }
  write_cylinder_round (v1, v2, current_state->stickradius, FALSE);
}


/*------------------------------------------------------------*/
void
pov_strand (void)
{
  int slot;
  int thickness = current_state->strandthickness >= 0.01;
  strand_segment *ss1, *ss2;

  mesh_output=1;

  write_material();
				/* strand face 1 */
  if (! current_state->colourparts)
    convert_colour (&(current_state->planecolour));

  for (slot = 1; slot < strand_segment_count - 3; slot++) {
    ss1 = strand_segments + slot - 1;
    ss2 = ss1 + 1;
    if (current_state->colourparts) convert_colour (&(ss1->c));
    write_norm_triangle (&(ss1->p1), &(ss1->p4), &(ss2->p1), 
			 &(ss1->n1), &(ss1->n1), &(ss2->n1));
    write_norm_triangle (&(ss1->p4), &(ss2->p1), &(ss2->p4), 
			 &(ss1->n1), &(ss2->n1), &(ss2->n1));
  }
				/* arrow face 1, high */
  ss1 = strand_segments + strand_segment_count - 3;
  ss2 = ss1 + 1;
  write_norm_triangle (&(ss1->p1), &(ss1->p4), &(ss2->p1), 
		       &(ss1->n1), &(ss1->n1), &(ss2->n1));
  write_norm_triangle (&(ss1->p4), &(ss2->p1), &(ss2->p4), 
		       &(ss1->n1), &(ss2->n1), &(ss2->n1));

				/* arrow face 2, high */
  ss1 = strand_segments + strand_segment_count - 2;
  ss2 = ss1 + 1;
  if (current_state->colourparts) convert_colour (&(ss1->c));
  write_norm_triangle (&(ss1->p1), &(ss1->p4), &(ss2->p1),
		       &(ss1->n1), &(ss1->n1), &(ss2->n1));

  if (thickness) {
    vector3 dir1, dir2, normal;
				/* strand face 2 */
    for (slot = 1; slot < strand_segment_count - 3; slot++) {
      ss1 = strand_segments + slot - 1;
      ss2 = ss1 + 1;
      if (current_state->colourparts) convert_colour (&(ss1->c));
      write_norm_triangle (&(ss1->p2), &(ss1->p3), &(ss2->p2), 
			   &(ss1->n3), &(ss1->n3), &(ss2->n3));
      write_norm_triangle (&(ss1->p3), &(ss2->p2), &(ss2->p3),
			   &(ss1->n3), &(ss2->n3), &(ss2->n3));
    }
				/* arrow face 1, low */
    ss1 = strand_segments + strand_segment_count - 3;
    ss2 = ss1 + 1;
    write_norm_triangle (&(ss1->p2), &(ss1->p3), &(ss2->p2),
			 &(ss1->n3), &(ss1->n3), &(ss2->n3));
    write_norm_triangle (&(ss1->p3), &(ss2->p2), &(ss2->p3),
			 &(ss1->n3), &(ss2->n3), &(ss2->n3));
				/* arrow face 2, low */
    ss1 = strand_segments + strand_segment_count - 2;
    ss2 = ss1 + 1;
    if (current_state->colourparts) convert_colour (&(ss1->c));
    write_norm_triangle (&(ss1->p2), &(ss1->p3), &(ss2->p2),
			 &(ss1->n3), &(ss1->n3), &(ss2->n2));
				/* strand side 1 */

    if (!current_state->colourparts)
      convert_colour (&(current_state->plane2colour));

    for (slot = 1; slot < strand_segment_count - 3; slot++) {
      ss1 = strand_segments + slot - 1;
      ss2 = ss1 + 1;
      if (current_state->colourparts) convert_colour (&(ss1->c));
      write_norm_triangle (&(ss1->p2), &(ss1->p1), &(ss2->p2), 
			   &(ss1->n2), &(ss1->n2), &(ss2->n2));
      write_norm_triangle (&(ss1->p1), &(ss2->p2), &(ss2->p1),
			   &(ss1->n2), &(ss2->n2), &(ss2->n2));
    }
				/* strand side 2 */
    for (slot = 1; slot < strand_segment_count - 3; slot++) {
      ss1 = strand_segments + slot - 1;
      ss2 = ss1 + 1;
      if (current_state->colourparts) convert_colour (&(ss1->c));
      write_norm_triangle (&(ss1->p3), &(ss1->p4), &(ss2->p3),
			   &(ss1->n4), &(ss1->n4), &(ss2->n4));
      write_norm_triangle (&(ss1->p4), &(ss2->p3), &(ss2->p4),
			   &(ss1->n4), &(ss2->n4), &(ss2->n4));
    }
				/* arrow base */
    ss1 = strand_segments + strand_segment_count - 3;
    ss2 = ss1 - 1;

    v3_difference (&dir1, &(ss1->p3), &(ss1->p1));
    v3_difference (&dir2, &(ss1->p4), &(ss1->p2));
    v3_cross_product (&normal, &dir1, &dir2);
    v3_normalize (&normal);

    if (current_state->colourparts) convert_colour (&(ss2->c));
    write_norm_triangle (&(ss1->p1), &(ss1->p2), &(ss2->p1),
			 &normal, &normal, &normal);
    write_norm_triangle (&(ss2->p1), &(ss1->p2), &(ss2->p2),
			 &normal, &normal, &normal);
    write_norm_triangle (&(ss1->p3), &(ss1->p4), &(ss2->p3),
			 &normal, &normal, &normal);
    write_norm_triangle (&(ss2->p3), &(ss1->p4), &(ss2->p4),
			 &normal, &normal, &normal);

    ss1 = strand_segments;	/* strand base */
    v3_difference (&dir1, &(ss1->p3), &(ss1->p2));
    v3_difference (&dir2, &(ss1->p1), &(ss1->p2));
    v3_cross_product (&normal, &dir1, &dir2);
    v3_normalize (&normal);


    if (current_state->colourparts) convert_colour (&(ss1->c));
    write_norm_triangle (&(ss1->p1), &(ss1->p2), &(ss1->p3),
			 &normal, &normal, &normal);
    write_norm_triangle (&(ss1->p3), &(ss1->p4), &(ss1->p1),
			 &normal, &normal, &normal);


				/* arrow side 1 */
    ss1 = strand_segments + strand_segment_count - 3;
    ss2 = ss1 + 1;

    v3_difference (&dir1, &(ss2->p1), &(ss1->p2));
    v3_difference (&dir2, &(ss2->p2), &(ss1->p1));
    v3_cross_product (&normal, &dir1, &dir2);
    v3_normalize (&normal);

    if (current_state->colourparts) convert_colour (&(ss1->c));
    write_norm_triangle (&(ss1->p2), &(ss1->p1), &(ss2->p2),
			 &normal, &normal, &normal);
    write_norm_triangle (&(ss1->p1), &(ss2->p2), &(ss2->p1),
			 &normal, &normal, &normal);
    ss1 = ss2;
    ss2++;

    v3_difference (&dir1, &(ss2->p1), &(ss1->p2));
    v3_difference (&dir2, &(ss2->p2), &(ss1->p1));
    v3_cross_product (&normal, &dir1, &dir2);
    v3_normalize (&normal);

    if (current_state->colourparts) convert_colour (&(ss1->c));
    write_norm_triangle (&(ss1->p2), &(ss1->p1), &(ss2->p2),
			 &normal, &normal, &normal);
    write_norm_triangle (&(ss1->p1), &(ss2->p2), &(ss2->p1),
			 &normal, &normal, &normal);

				/* arrow side 2 */
    ss1 = strand_segments + strand_segment_count - 3;
    ss2 = ss1 + 1;

    v3_difference (&dir1, &(ss2->p3), &(ss1->p4));
    v3_difference (&dir2, &(ss2->p4), &(ss1->p3));
    v3_cross_product (&normal, &dir1, &dir2);
    v3_normalize (&normal);

    if (current_state->colourparts) convert_colour (&(ss1->c));
    write_norm_triangle (&(ss1->p3), &(ss1->p4), &(ss2->p3),
			 &normal, &normal, &normal);
    write_norm_triangle (&(ss1->p4), &(ss2->p3), &(ss2->p4),
			 &normal, &normal, &normal);
    ss1 = ss2;
    ss2++;

    v3_difference (&dir1, &(ss2->p2), &(ss1->p4));
    v3_difference (&dir2, &(ss2->p1), &(ss1->p3));
    v3_cross_product (&normal, &dir1, &dir2);
    v3_normalize (&normal);

    if (current_state->colourparts) convert_colour (&(ss1->c));
    write_norm_triangle (&(ss1->p3), &(ss1->p4), &(ss2->p2),
			 &normal, &normal, &normal);
    write_norm_triangle (&(ss1->p4), &(ss2->p2), &(ss2->p1),
			 &normal, &normal, &normal);
  }

  if (triangle_count > 0)
    pov_fin_object();
}


/*------------------------------------------------------------*/
static void
convert_vector_colour (vector3 *v)
{
  assert (v);

  pstate.rgb.x = v->x;
  pstate.rgb.y = v->y; 
  pstate.rgb.z = v->z;
}

/*------------------------------------------------------------*/
void
pov_start_object(void)
{
  if (mesh_output)
    start_group(GR_MESH);
  else
    start_group(GR_GROUP_OR_MERGE);
}

/*------------------------------------------------------------*/
void
pov_fin_object(void)
{
  if (mesh_output){
    end_group(GR_MESH);
    if (message_mode)
      fprintf (stderr, "%i triangles in mesh output\n", triangle_count);
    triangle_count = 0;
    mesh_output = 0;
  }
  else
    end_group(GR_GROUP_OR_MERGE);
}

/*------------------------------------------------------------*/
void
pov_object (int code, vector3 *triplets, int count)
{
  int slot;
  vector3 *v;
  vector3 normal;
  double radius;

  assert (triplets);
  assert (count > 0);

  write_material();

  switch (code) {

  case OBJ_POINTS:
    radius = LINEWIDTH_FACTOR * current_state->linewidth;
    if (radius < LINEWIDTH_MINIMUM) radius = LINEWIDTH_MINIMUM;
    convert_colour (&(current_state->linecolour));
    for (slot = 0; slot < count; slot++) {
      write_sphere (triplets + slot, radius);
    }
    break;

  case OBJ_POINTS_COLOURS:
    radius = LINEWIDTH_FACTOR * current_state->linewidth;
    if (radius < LINEWIDTH_MINIMUM) radius = LINEWIDTH_MINIMUM;
    for (slot = 0; slot < count; slot += 2) {
      convert_vector_colour (triplets + slot + 1);
      write_sphere (triplets + slot, radius);
    }
    break;

  case OBJ_LINES:
    convert_colour (&(current_state->linecolour));
    for (slot = 1; slot < count; slot++) {
      write_line (triplets + slot - 1, triplets + slot);
    }
    break;

  case OBJ_LINES_COLOURS:
    for (slot = 2; slot < count; slot += 2) {
      v = triplets + slot;
      convert_vector_colour (v - 1);
      write_line (v - 2, v);
    }
    break;

  case OBJ_TRIANGLES:
    mesh_output=1;
    convert_colour (&(current_state->planecolour));
    for (slot = 0; slot < count; slot += 3) {
      v = triplets + slot;
      write_triangle (v, v + 1, v + 2);
    }
    break;

  case OBJ_TRIANGLES_COLOURS:
    mesh_output=1;
    for (slot = 0; slot < count; slot += 6) {
      v = triplets + slot;
      convert_vector_colour (v + 1);
      write_triangle (v, v + 2, v + 4);
    }
    break;

  case OBJ_TRIANGLES_NORMALS:
    mesh_output=1;
    convert_colour (&(current_state->planecolour));
    for (slot = 0; slot < count; slot += 6) {
      v = triplets + slot;
      write_norm_triangle (v, v + 2, v + 4,
			   v + 1, v + 3, v + 5);
    }
    break;

  case OBJ_TRIANGLES_NORMALS_COLOURS:
    mesh_output=1;
    for (slot = 0; slot < count; slot += 9) {
      v = triplets + slot;
      convert_vector_colour (v + 2);
      write_norm_triangle (v, v + 3, v + 6, 
			   v + 1, v + 4, v + 7);
    }
    break;

  case OBJ_STRIP:
    mesh_output=1;
    convert_colour (&(current_state->planecolour));
    for (slot = 2; slot < count; slot++) {
      v = triplets + slot;
      v3_triangle_normal (&normal, v - 2, v - 1, v);
      write_norm_triangle (v - 2, v - 1, v, 
			   &normal, &normal, &normal);
    }
    break;

  case OBJ_STRIP_COLOURS:
    mesh_output=1;
    for (slot = 4; slot < count; slot += 2) {
      v = triplets + slot;
      convert_vector_colour (v + 1);
      v3_triangle_normal (&normal, v - 4, v - 2, v);
      write_norm_triangle (v - 4, v - 2, v, 
			   &normal, &normal, &normal);
    }
    break;

  case OBJ_STRIP_NORMALS:
    mesh_output=1;
    convert_colour (&(current_state->planecolour));
    for (slot = 4; slot < count; slot += 2) {
      v = triplets + slot;
      write_norm_triangle (v - 4, v - 2, v, 
			   v - 3, v - 1, v + 1);
    }
    break;

  case OBJ_STRIP_NORMALS_COLOURS:
    mesh_output=1;
    for (slot = 6; slot < count; slot += 3) {
      v = triplets + slot;
      convert_vector_colour (v + 2);
      write_norm_triangle (v - 6, v - 3, v,
			   v - 5, v - 2, v + 1);
    }
    break;

  }
}
