/* surfmap.c
 *
 * read in GRASP surface map format files
 * for display in Molscript
 * bits and pieces of code from pov4grasp (by Nicolas Calimet):
 * http://pov4grasp.free.fr/
 *
 * 4/9/02 - TimF - first attempts
 * 5/20/02 - TimF - add MSMS support
 * 6/10/02 - TimF - add extra MSMS features
 * 3/9/03 - TimF - add support for mesh2 output
 * 4/12/04 - TimF - use GTS for output
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gts.h>

#include "clib/aa_lookup.h"
#include "clib/colour.h"
#include "clib/str_utils.h"
#include "clib/extent3d.h"
#include "clib/matrix3.h"
#include "clib/mol3d.h"
#include "clib/vector3.h"

#include "global.h"
#include "graphics.h"
#include "map_utils.h"
#include "segment.h"
#include "select.h"
#include "state.h"
#include "surfmap.h"
#include "vertex_class.h"

FILE *gfp, *gfp2;

void generate_gtssurface(long *trinum, vector3 *trivert, vector3 *normvert, float *colorprop){
  float cprop;
  long i, j;
  GtsFace *f;
  GtsEdge *e[3];
  GtsVertex2 *v[3];

  curmap->surface = gts_surface_new(gts_surface_class(), gts_face_class(), gts_edge_class(), GTS_VERTEX_CLASS(gts_vertex2_class()));

  for (i=0; i<(curmap->ntris)*3; i+=3){
    /* skip nonsense triangles (happens w/MSMS) */
    if ((v3_distance(&trivert[trinum[i]], &trivert[trinum[i+1]]) <= 0.0) || (v3_distance(&trivert[trinum[i]], &trivert[trinum[i+2]]) <= 0.0) ||
	(v3_distance(&trivert[trinum[i+1]], &trivert[trinum[i+2]]) <= 0.0))
      continue;

    /* setup vertices */
    for (j=0; j<3; j++){
      cprop = (curmap->drawflag == 0) ? 0.0 : colorprop[trinum[i+j]];
      v[j] = gts_vertex2_new(gts_vertex2_class(),
			     trivert[trinum[i+j]].x, trivert[trinum[i+j]].y, trivert[trinum[i+j]].z,
			     normvert[trinum[i+j]].x, normvert[trinum[i+j]].y, normvert[trinum[i+j]].z,
			     0.0, 0.0, 0.0,
			     cprop);
    }

    /* make sure the order of vertices is correct... */
    e[0] = gts_edge_new(gts_edge_class(), GTS_VERTEX(v[0]), GTS_VERTEX(v[1]));
    e[1] = gts_edge_new(gts_edge_class(), GTS_VERTEX(v[1]), GTS_VERTEX(v[2]));
    e[2] = gts_edge_new(gts_edge_class(), GTS_VERTEX(v[0]), GTS_VERTEX(v[2]));

    f = gts_face_new(gts_face_class(), e[0], e[1], e[2]);

    gts_surface_add_face(curmap->surface, f);
  }  
}

/* read in map */
void open_srfmap(char *filename){
  int sel, atom_count, slot, format=NONE;
  char buf[MAXLINE], newfile[MAXLINE];
  vector3 low, high;
  at3d **atoms;
  GtsFile *tmpgfile;
  GtsSurface *tmpgsurf;

  sel = count_atom_selections();
  atom_count = 0;

  /* molscript initialization */
  if (sel){
    assert (sel == 1);

    ext3d_push();
    ext3d_initialize();

    atoms = select_atom_list (&atom_count);
    if (message_mode)
      fprintf (stderr, "%i atoms selected for map reading\n", atom_count);
 
    for (slot = 0; slot < atom_count; slot++)
      ext3d_update (&(atoms[slot]->xyz), 0.0);

    free(atoms);

    ext3d_get_low_high(&low, &high);
    ext3d_pop (FALSE);
    assert (count_atom_selections() == 0);
  }

  /* determine file type */
  if ((gfp = fopen(filename, "rb")) != NULL){
    fseek(gfp, 4, SEEK_SET);
    fread(buf, 80, sizeof(char), gfp);
    if (!strncmp(buf, "format=2", 8)){
      fprintf(stderr, "GRASP surface format 2 found for %s\n", filename);
      format=GRASP;
    }
    rewind(gfp);

    /* not GRASP? */
    if (format == NONE){
      tmpgfile = gts_file_new(gfp);
      tmpgsurf = gts_surface_new(gts_surface_class(), gts_face_class(), gts_edge_class(), GTS_VERTEX_CLASS(gts_vertex2_class()));
      if (!gts_surface_read(tmpgsurf, tmpgfile)){
	fprintf(stderr, "GTS surface found for %s\n", filename);
	format=GTS;
      }
      gts_object_destroy(tmpgsurf);
      gts_file_destroy(tmpgfile);
      rewind(gfp);
    }
  }
  else{
    /* MSMS? */
    sprintf(newfile, "%s.vert", filename);
    if ((gfp = gfopen(newfile, "rb")) != NULL){
      sprintf(newfile, "%s.face", filename);
      if ((gfp2 = gfopen(newfile, "rb")) != NULL){
	fprintf(stderr, "MSMS surface assumed for file %s\n", filename);
	format=MSMS;
      }
      else{
	exit_on_error = TRUE;
	yyerror("surface map file not found");
      }
    }
  }
  if (gfp == NULL){
    fprintf(stderr, "GRASP surface %s (or its MSMS equivalent: %s.vert and %s.face) not found\n", filename, filename, filename);
    return;
  }

  /* read and close the file */
  if (sel && atom_count){
    fprintf(stderr, "read %s with low and high limits\n", filename);
    read_srfheader(format, &low, &high);
  }
  else{
    fprintf(stderr, "read %s with no limits\n", filename);
    read_srfheader(format, NULL, NULL);
  }

  if (format != GTS)
    read_srfdata();
  fclose(gfp);
  if (format == MSMS)
    fclose(gfp2);
}

/* read GRASP file data */
void read_srfdata(void){
  short trifacecode, vertfacecode;
  long int i, trifacenum, vertfacenum, csphere;
  char buf[MAXLINE], closeres[MAXLINE];

  long *trinum;
  vector3 *trivert;
  vector3 *normvert;
  float *colorprop;

  /* allocate memory for the tris and verts */
  if ((trivert = calloc(curmap->nverts, sizeof(vector3))) == NULL){
    perror("calloc");
    exit(1);
  }
  if ((normvert = calloc(curmap->nverts, sizeof(vector3))) == NULL){
    perror("calloc");
    exit(1);
  }
  if (curmap->drawflag != 0){
    if ((colorprop = calloc(curmap->nverts, sizeof(float))) == NULL){
      perror("calloc");
      exit(1);
    }
  }
  if ((trinum = calloc((curmap->ntris*3), sizeof(long))) == NULL){
    perror("calloc");
    exit(1);
  }

  if (curmap->format == GRASP){

    /* jump to verts */
    fseek(gfp, 8, SEEK_CUR);
    for (i=0; i<curmap->nverts; i++){
      /* read in each vertex, byte swap if necessary */
      fread((char *)&buf, 1, sizeof(float), gfp);
      #ifndef WORDS_BIGENDIAN
      swap((char *)&buf, sizeof(float));
      #endif
      trivert[i].x = *(float *)buf;

      fread((char *)&buf, 1, sizeof(float), gfp);
      #ifndef WORDS_BIGENDIAN
      swap((char *)&buf, sizeof(float));
      #endif
      trivert[i].y = *(float *)buf;

      fread((char *)&buf, 1, sizeof(float), gfp);
      #ifndef WORDS_BIGENDIAN
      swap((char *)&buf, sizeof(float));
      #endif
      trivert[i].z = *(float *)buf;
    }

    /* skip the accessibles */
    fseek(gfp, 8, SEEK_CUR);
    fseek(gfp, curmap->nverts*12, SEEK_CUR);

    /* jump to normals */
    fseek(gfp, 8, SEEK_CUR);
    for (i=0; i<curmap->nverts; i++){
      /* now read in normals */
      fread((char *)&buf, 1, sizeof(float), gfp);
      #ifndef WORDS_BIGENDIAN
      swap((char *)&buf, sizeof(float));
      #endif
      normvert[i].x = *(float *)buf;

      fread((char *)&buf, 1, sizeof(float), gfp);
      #ifndef WORDS_BIGENDIAN
      swap((char *)&buf, sizeof(float));
      #endif
      normvert[i].y = *(float *)buf;

      fread((char *)&buf, 1, sizeof(float), gfp);
      #ifndef WORDS_BIGENDIAN
      swap((char *)&buf, sizeof(float));
      #endif
      normvert[i].z = *(float *)buf;
    }

    /* now read in triangles */
    fseek(gfp, 8, SEEK_CUR);
    for (i=0; i<(curmap->ntris)*3; i++){
      /* read triangle number */
      fread((char *)&buf, 1, sizeof(long), gfp);
      #ifndef WORDS_BIGENDIAN
      swap((char *)&buf, sizeof(long));
      #endif
      trinum[i] = *(long *)buf - 1;
    }

    /* if we're reading more stuff, do it */
    if (curmap->drawflag != 0){
      curmap->collow = 0.0;
      curmap->colhigh = 0.0;

      /* jump to the appropriate position */
      fseek(gfp, 8, SEEK_CUR);
      for (i=1; i<curmap->drawflag; i*=2){
	if (curmap->flag & i)
	  fseek (gfp, curmap->nverts*4 + 8, SEEK_CUR);
      }
      /* now read in colour data */
      for (i=0; i<curmap->nverts; i++){
	fread((char *)&buf, 1, sizeof(float), gfp);
        #ifndef WORDS_BIGENDIAN
	swap((char *)&buf, sizeof(float));
	#endif
	colorprop[i] = *(float *)buf;

	if (colorprop[i] > curmap->colhigh)
	  curmap->colhigh = colorprop[i];
	if (colorprop[i] < curmap->collow)
	  curmap->collow = colorprop[i];
      }
      curmap->colmid = (curmap->colhigh + curmap->collow) / 2.0;
      fprintf(stderr, "GRASP colour low: %f mid: %f high: %f\n", curmap->collow, curmap->colmid, curmap->colhigh);
    }
  }
  else{
    /* MSMS data */
    for (i=0; i<curmap->nverts; i++){
      if ((fgets(buf, MAXLINE, gfp)) == NULL){
	fprintf(stderr, "MSMS EOF reached prematurely\n");
	return;
      }
      /* read in verts */
      sscanf(buf, "%lf %lf %lf %lf %lf %lf %ld %ld %hd %s", &(trivert[i].x), &(trivert[i].y), &(trivert[i].z), &(normvert[i].x), &(normvert[i].y), &(normvert[i].z), &vertfacenum, &csphere, &vertfacecode, closeres);

      /* convert residue type to a color property */
      if (curmap->drawflag != 0)
	colorprop[i] = aa_ordinal_str(closeres);
    }

    for (i=0; i<(curmap->ntris)*3; i+=3){
      if ((fgets(buf, MAXLINE, gfp2)) == NULL){
	fprintf(stderr, "MSMS EOF reached prematurely\n");
	return;
      }
      /* read in tris */
      sscanf(buf, "%ld %ld %ld %hd %ld", &(trinum[i]), &(trinum[i+1]), &(trinum[i+2]), &trifacecode, &trifacenum);
      /* MSMS triangle index starts at 1 */
      trinum[i] -= 1;
      trinum[i+1] -= 1;
      trinum[i+2] -= 1;
    }
  }

  /* we got what we need - now put it in GTS format */
  generate_gtssurface(trinum, trivert, normvert, colorprop);
  free(trinum);
  free(trivert);
  free(normvert);
  if (curmap->drawflag != 0)
    free(colorprop);
}

void read_srfheader(int format, vector3 *low, vector3 *high){
  int i, grid;
  int dr, dg, db;
  int bytes_read;
  char buf[MAXLINE];
  GtsFile *tmpgfile;

  curmap->format = format;

  if (format == GRASP){
    /* read in full header */
    fseek(gfp, 4, SEEK_SET);
    fread(buf, 80, sizeof(char), gfp);
    /* format type checks here */

    /* line 2 */
    fseek(gfp, 8, SEEK_CUR);
    bytes_read = fread(buf, 80, sizeof(char), gfp);

    /* line 3 */
    fseek(gfp, 8, SEEK_CUR);
    bytes_read = fread(buf, 80, sizeof(char), gfp);

    curmap->flag=0;
    i=0;
    if (buf[0] == ',')
      i++;
    while ((i<80) && (buf[i] != ' ')){
      if (! strncmp(buf+i, "potentials", 10))
	curmap->flag |= POTENTIALS;
      if (! strncmp(buf+i, "curvature", 9))
	curmap->flag |= CURVATURES;
      if (! strncmp(buf+i, "distances", 9))
	curmap->flag |= DISTANCES;
      if (! strncmp(buf+i, "gproperty", 9))
	curmap->flag |= PROPERTY_1;
      if (! strncmp(buf+i, "g2property", 10))
	curmap->flag |= PROPERTY_2;
      if (! strncmp(buf+i, "vertexcolor", 11))
	curmap->flag |= VERTEXCOLOR;

      i++;
    }

    #ifdef EDEBUG
    fprintf(stderr, "flag: %d\n", curmap->flag);
    #endif

    /* line 4 */
    fseek(gfp, 8, SEEK_CUR);
    bytes_read = fread(buf, 80, sizeof(char), gfp);
    sscanf(buf, "%ld%ld%d%f", &(curmap->nverts), &(curmap->ntris), &grid, &(curmap->rspacing));
    fprintf(stderr, "GRASP nverts: %ld ntris: %ld grid: %d recip. spacing: %f\n", curmap->nverts, curmap->ntris, grid, curmap->rspacing);

    /* line 5 */
    fseek(gfp, 8, SEEK_CUR);
    bytes_read = fread(buf, 80, sizeof(char), gfp);
    sscanf(buf, "%lf%lf%lf", &(curmap->mid.x), &(curmap->mid.y), &(curmap->mid.z));
    fprintf(stderr, "GRASP midpoint x:%f y:%f z:%f\n", curmap->mid.x, curmap->mid.y, curmap->mid.z);

    /* setup discrete coloring scheme */
    if (curmap->coldef){
      if ((gfp2 = gfopen(curmap->coldef, "rb")) != NULL){
	fprintf(stderr, "GRASP disrete color file %s read in\n", curmap->coldef);
	for (i=0; i<100; i++){
	  fgets(buf, 80, gfp2);
	  sscanf(buf, "%d%d%d", &db, &dg, &dr);
	  fprintf(stderr, "%2.1f %2.1f %2.1f\n", ((double) dr) / 255.0, 
		  ((double) dg) / 255.0,
		  ((double) db) / 255.0);
	  colour_set_rgb(&curmap->comp[i], ((double) dr) / 255.0, 
			 ((double) dg) / 255.0,
			 ((double) db) / 255.0);
	}
      }
    }
    else{
      /* order: white, red, green, blue, magenta, cyan, yellow, grey, navy */
      colour_set_rgb(&curmap->comp[0], 1.0, 1.0, 1.0);
      colour_set_rgb(&curmap->comp[1], 1.0, 0.0, 0.0);
      colour_set_rgb(&curmap->comp[2], 0.0, 1.0, 0.0);
      colour_set_rgb(&curmap->comp[3], 0.0, 0.0, 1.0);
      colour_set_rgb(&curmap->comp[4], 1.0, 0.0, 1.0);
      colour_set_rgb(&curmap->comp[5], 0.0, 1.0, 1.0);
      colour_set_rgb(&curmap->comp[6], 1.0, 1.0, 0.0);
      colour_set_rgb(&curmap->comp[7], 0.5, 0.5, 0.5);
      colour_set_rgb(&curmap->comp[8], 0.2, 0.4, 0.6);
    }
  }
  else if (format == GTS){
    tmpgfile = gts_file_new(gfp);
    curmap->surface = gts_surface_new(gts_surface_class(), gts_face_class(), gts_edge_class(), GTS_VERTEX_CLASS(gts_vertex2_class()));
    if (gts_surface_read(curmap->surface, tmpgfile)){
      fprintf(stderr, "GTS surface error: %s\n", tmpgfile->error);
    }
    gts_file_destroy(tmpgfile);
  }
  else{
    /* read init values from MSMS file */
    fgets(buf, MAXLINE, gfp);
    fgets(buf, MAXLINE, gfp);
    fgets(buf, MAXLINE, gfp);
    sscanf(buf, "%ld %ld %f %f\n", &(curmap->nverts), &(curmap->nspheres), &(curmap->density), &(curmap->pradius));

    fgets(buf, MAXLINE, gfp2);
    fgets(buf, MAXLINE, gfp2);
    fgets(buf, MAXLINE, gfp2);
    sscanf(buf, "%ld", &(curmap->ntris));

    fprintf(stderr, "MSMS ntris: %ld nverts: %ld nspheres: %ld density: %f probe_radius: %f\n", curmap->ntris, curmap->nverts, curmap->nspheres, curmap->density, curmap->pradius);
  }
}
