#include "vertex_class.h"

#include <stdio.h>
#include <stdlib.h>

#include <clib/matrix3.h>
#include <clib/mol3d.h>
#include <clib/vector3.h>

#include "map_utils.h"

map_ptr map1, curmap;

/* GTS vertex class with inheritance properties */

static void vertex2_destroy (GtsObject * object){
  (* GTS_OBJECT_CLASS (gts_vertex_class())->parent_class->destroy) (object);
}

static void vertex2_read (GtsObject ** object, GtsFile * fp){
  gint i;

  (* GTS_OBJECT_CLASS (gts_vertex_class())->parent_class->read) (object, fp);
  if (fp->type == GTS_ERROR)
    return;

  /* extra vertex info */

  /*
   * warning - vertices not transformed/normalized
   * by molscript matrices (should we allow an option for this?)
   */
  gts_file_first_token_after (fp, '\n');
  for (i=0; i<3; i++){
    if (fp->type != GTS_FLOAT && fp->type != GTS_INT){
      gts_file_error (fp, "expecting a float/int (normal vector %d)", i);
      return;
    }
    GTS_VERTEX2(*object)->norm[i] = atof(fp->token->str);
    gts_file_next_token (fp);
  }

  gts_file_first_token_after (fp, '\n');
  for (i=0; i<3; i++){
    if (fp->type != GTS_FLOAT && fp->type != GTS_INT){
      gts_file_error (fp, "expecting a float/int (color vector %d)", i);
      return;
    }
    GTS_VERTEX2(*object)->col[i] = atof(fp->token->str);
    gts_file_next_token (fp);
  }
}

static void vertex2_write (GtsObject * object, FILE * fp){
  vector3 u;

  /* we want to transform the point, so don't call parent class */
  /* (* GTS_OBJECT_CLASS (gts_vertex_class())->parent_class->write) (object, fp); */
  /* extra vertex info */
  u.x = GTS_POINT(object)->x;
  u.y = GTS_POINT(object)->y;
  u.z = GTS_POINT(object)->z;
  matrix3_transform(&u, curmap->f_t_c);
  fprintf(fp, "%g %g %g", u.x, u.y, u.z);

  /* normal */
  u.x = GTS_VERTEX2(object)->norm[0];
  u.y = GTS_VERTEX2(object)->norm[1];
  u.z = GTS_VERTEX2(object)->norm[2];
  matrix3_rotate(&u, curmap->f_t_c);
  v3_normalize(&u);
  fprintf(fp, "\n%g %g %g", u.x, u.y, u.z);

  /* color (0-1 scale) */
  fprintf(fp, "\n%g %g %g", GTS_VERTEX2(object)->col[0], GTS_VERTEX2(object)->col[1], GTS_VERTEX2(object)->col[2]);
}

void vertex2_class_init (GtsObjectClass * klass){
  GTS_OBJECT_CLASS(klass)->destroy = vertex2_destroy;
  GTS_OBJECT_CLASS(klass)->read = vertex2_read;
  GTS_OBJECT_CLASS(klass)->write = vertex2_write;
}

GtsVertex2Class *gts_vertex2_class( void ) {
  static GtsVertex2Class *klass = NULL;

  if (klass == NULL){
    GtsObjectClassInfo vertex2_info = {
      "GtsVertex2",
      sizeof(GtsVertex2), sizeof(GtsVertex2Class),
      (GtsObjectClassInitFunc) vertex2_class_init,
      (GtsObjectInitFunc) NULL,
      (GtsArgSetFunc) NULL,
      (GtsArgGetFunc) NULL
    };

    klass = gts_object_class_new(gts_object_class(), &vertex2_info);
  }
  return(klass);
}

GtsVertex2 *gts_vertex2_new (GtsVertex2Class *klass, double x, double y, double z,
			     double nx, double ny, double nz,
			     double r, double g, double b,
			     float cprop){
  GtsVertex2 *object;

  object = GTS_VERTEX2(gts_object_new(GTS_OBJECT_CLASS(klass)));
  gts_point_set(GTS_POINT(object), x, y, z);

  object->draw = TRUE;
  object->norm[0] = nx;
  object->norm[1] = ny;
  object->norm[2] = nz;
  object->col[0] = r;
  object->col[1] = g;
  object->col[2] = b;
  object->cprop = cprop;

  return(object);
}
